#!/usr/bin/python3

import os

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gio", "2.0")
gi.require_version("GLib", "2.0")
from gi.repository import Gtk, Gio, GLib

from bin.SettingsWidgets import SidePage
from xapp.SettingsWidgets import SettingsStack, SettingsPage, SettingsSection, SettingsWidget, SettingsLabel


BOLT_BUS_NAME = "org.freedesktop.bolt"
BOLT_OBJECT_PATH = "/org/freedesktop/bolt"


def build_detail_row(key, value):
    row = SettingsWidget()
    labelKey = Gtk.Label(label=key)
    labelKey.get_style_context().add_class("dim-label")
    row.pack_start(labelKey, False, False, 0)
    if isinstance(value, Gtk.Label):
        labelValue = value
    else:
        labelValue = Gtk.Label(label=str(value))
    labelValue.set_selectable(True)
    labelValue.set_line_wrap(True)        
    row.pack_end(labelValue, False, False, 0)
    return row

def format_generation(gen):
    if gen in (1,2,3):
        return f'Thunderbolt {gen}'
    elif gen == 4:
        return 'USB4'
    raise ValueError("undefined thunderbolt generation")


class DBusProps:
    pass


class DBusProxy:
    def __init__(self, name, object_path, interface_name):
        # Save the dbus info
        self.name = name
        self.object_path = object_path
        self.interface_name = interface_name

        # Callback for when properties are changed
        self.on_property_changed = None
        self.on_property_invalidated = None

        # Get the initial properties for this DBus Proxy
        var = Gio.DBusConnection.call_sync(
            Gio.bus_get_sync(Gio.BusType.SYSTEM),
            name,
            object_path,
            "org.freedesktop.DBus.Properties",
            "GetAll",
            GLib.Variant("(s)", (interface_name,)),
            None,
            0,
            -1,
            None)

        self.props = DBusProps()
        (props,) = var.unpack()
        for key, value in props.items():
            setattr(self.props, key, value)

        # Main proxy for DBus Object
        self._proxy = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SYSTEM,
            Gio.DBusProxyFlags.NONE,
            None,
            name,
            object_path,
            interface_name,
            None)

        # Register for future changes to properties
        self._proxy.connect("g-properties-changed", self._on_g_properties_changed)

    def _on_g_properties_changed(self, proxy, changed, invalidated):
        for key, value in changed.unpack().items():
            setattr(self.props, key, value)
            if self.on_property_changed:
                self.on_property_changed(key, value)
        for key in invalidated:
            delattr(self.props, key)
            if self.on_property_invalidated:
                self.on_property_invalidated(key)


class BoltManagerProxy(DBusProxy):
    def __init__(self):
        # Perform parent initialization
        super().__init__(
            BOLT_BUS_NAME,
            BOLT_OBJECT_PATH,
            "org.freedesktop.bolt1.Manager"
            )

        # Callbacks
        self.on_device_added = None
        self.on_device_removed = None
        
        # Connect to g-signal for event handling
        self._proxy.connect('g-signal', self._on_g_signal)

    def _on_g_signal(self, proxy, sender, signal, parameters):
        if signal == "DeviceAdded" and self.on_device_added:
            (obj_path,) = parameters.unpack()
            self.on_device_added(obj_path)
        elif signal == "DeviceRemoved" and self.on_device_removed:
            (obj_path,) = parameters.unpack()
            self.on_device_removed(obj_path)

    def list_devices(self):
        return self._proxy.ListDevices()

    def enroll_device(self, uid):
        self._proxy.EnrollDevice('(sss)', uid, 'auto', '')

    def forget_device(self, uid):
        self._proxy.ForgetDevice('(s)', uid)


class BoltDeviceProxy(DBusProxy):
    def __init__(self, obj_path):
        super().__init__(
            BOLT_BUS_NAME,
            obj_path,
            "org.freedesktop.bolt1.Device"
            )

    def authorize(self):
        self._proxy.Authorize('(s)', 'auto')


class BoltSection(SettingsSection):

    def __init__(self, bolt_manager, bolt_device):
        self.bolt_manager = bolt_manager
        self.bolt_device = bolt_device
        self.bolt_device.on_property_changed = lambda k, v: self.refresh()
        super().__init__("{0} {1}".format(bolt_device.props.Vendor, bolt_device.props.Name))
        
        widget = SettingsWidget()
        self.status_label = SettingsLabel()
        widget.pack_start(self.status_label, False, False, 0)
        self.details_btn = Gtk.ToggleButton(label=_("Details"))
        self.details_btn.connect("toggled", lambda w: self.details_revealer.set_reveal_child(w.get_active()))
        self.auth_btn = Gtk.Button(label=_("Authorize"))
        self.auth_btn.connect("clicked", lambda w: self.bolt_device.authorize())
        self.trust_btn = Gtk.Button(label=_("Trust"))
        self.trust_btn.connect("clicked", self.on_trust_btn_clicked)
        button_box = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
        button_box.pack_start(self.details_btn, True, True, 0)
        button_box.pack_start(self.auth_btn, True, True, 0)
        button_box.pack_start(self.trust_btn, True, True, 0)
        button_box.set_layout(Gtk.ButtonBoxStyle.EXPAND)
        widget.pack_end(button_box, False, False, 0)
        self.add_row(widget)
        
        list_box = Gtk.ListBox()
        list_box.set_selection_mode(Gtk.SelectionMode.NONE)
        list_box.set_header_func(self.update_header)
        generation = self.bolt_device.props.Generation
        list_box.add(build_detail_row(_("Generation"), format_generation(generation)))
        self.details_bandwidth_label = Gtk.Label(label="-")
        list_box.add(build_detail_row(_("Bandwidth"), self.details_bandwidth_label))
        dev_type = self.bolt_device.props.Type
        list_box.add(build_detail_row(_("Type"), dev_type))
        uid = self.bolt_device.props.Uid
        list_box.add(build_detail_row("Uid", uid))
        self.details_revealer = self.add_reveal_row(list_box)

        self.refresh()

    def on_trust_btn_clicked(self, widget):
        uid = self.bolt_device.props.Uid
        stored = self.bolt_device.props.Stored
        if stored:
            self.bolt_manager.forget_device(uid)
        else:
            self.bolt_manager.enroll_device(uid)

    def update_header(self, row, before):
        if before:
            row.set_header(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))

    def refresh(self):
        status = self.bolt_device.props.Status
        stored = self.bolt_device.props.Stored
        link_speed = None
        if hasattr(self.bolt_device.props, "LinkSpeed"):
            link_speed = self.bolt_device.props.LinkSpeed

        # Update the status label
        text = _("Disconnected")
        if status in ("connected", "authorizing", "authorized"):
            text = _("Connected")
        if status == "authorized":
            text = text + " & " + _("Authorized")
        if stored:
            text = text + ", " + _("Trusted")
        self.status_label.set_label_text(text)

        # Update which buttons are active
        if status == "connected":
            self.auth_btn.set_sensitive(True)
            self.trust_btn.set_sensitive(False)
        else:
            self.auth_btn.set_sensitive(False)
            self.trust_btn.set_sensitive(True)

        # Update trust button label
        if stored:
            self.trust_btn.set_label(_("Forget"))
        else:
            self.trust_btn.set_label(_("Trust"))

        # Update bandwidth label
        if status == "disconnected":
            self.details_bandwidth_label.set_label("-")
        else:
            if link_speed:
                speed = link_speed['tx.speed']
                lanes = link_speed['tx.lanes']
                bandwidth = "{0} Gb/s ({1} {2} @ {3} Gb/s)".format(lanes * speed, lanes, _("lanes"), speed)
                self.details_bandwidth_label.set_label(bandwidth)
            else:
                self.details_bandwidth_label.set_label("-")


class Module:
    name = "thunderbolt"
    category = "hardware"
    comment = _("Manage Thunderbolt and USB4 devices")

    def __init__(self, content_box):
        keywords = _("thunderbolt, usb, docking, station, hub, dock")
        sidePage = SidePage("Thunderbolt", "cs-thunderbolt", keywords, content_box,
                            module=self)
        self.sidePage = sidePage
        self.bolt_manager = None
        self.bolt_devices = dict()

    def on_module_selected(self, check_again=False):
        # Check if thunderbolt is present
        thunderbolt_present = os.path.isdir("/sys/bus/thunderbolt")
        # Check if bolt is installed by finding 'boltctl'
        bolt_installed = GLib.find_program_in_path("boltctl")
        # Check if bolt is available on DBus
        boltd_alive = self.test_daemon_alive()

        # Check if we've already been loaded
        if not self.loaded:

            print("Loading Thunderbolt module")

            self.sidePage.stack = SettingsStack()
            self.sidePage.stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
            self.sidePage.add_widget(self.sidePage.stack)

            # Init the start page
            # Intentionally blank and only used during loading
            page = SettingsPage()
            self.sidePage.stack.add_named(page, "start")

            # Init the disabled page
            page = SettingsPage()
            self.sidePage.stack.add_named(page, "disabled")
            page.set_spacing(10)
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10, valign=Gtk.Align.START, margin_top=150)
            page.pack_start(box, True, True, 0)
            image = Gtk.Image(icon_name="xsi-dialog-warning-symbolic", icon_size=Gtk.IconSize.DIALOG)
            box.pack_start(image, False, False, 0)
            self.disabled_label = Gtk.Label(expand=True)
            box.pack_start(self.disabled_label, False, False, 0)
            self.disabled_retry_button = Gtk.Button(label=_("Check again"), no_show_all=True, halign=Gtk.Align.CENTER)
            self.disabled_retry_button.connect("clicked", self.disable_retry_on_clicked)
            box.pack_start(self.disabled_retry_button, False, False, 0)

            # Init the no devices page
            page = SettingsPage()
            self.sidePage.stack.add_named(page, "empty")
            page.set_spacing(10)
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10, valign=Gtk.Align.START, margin_top=150)
            page.pack_start(box, True, True, 0)
            label = Gtk.Label(label=_("No Thunderbolt or USB4 devices found."))
            box.pack_start(label, False, False, 0)

            # Init the settings page
            page = SettingsPage()
            self.sidePage.stack.add_named(page, "settings")
            page.set_spacing(10)

        show_disabled = False
        if not thunderbolt_present:
            text = _("Thunderbolt or USB4 is not detected on your system.")
            self.disabled_retry_button.set_visible(False)
            show_disabled = True
        elif not bolt_installed:
            text = _("The 'bolt' package must be installed to manage Thunderbolt and USB4 devices.")
            self.disabled_retry_button.set_visible(True)
            self.disabled_retry_button.set_sensitive(True)
            show_disabled = True
        elif not boltd_alive:
            text = _("The boltd service is not running.")
            self.disabled_retry_button.set_visible(True)
            self.disabled_retry_button.set_sensitive(True)
            show_disabled = True

        if show_disabled:
            self.reset()
            page = "disabled"
            self.disabled_label.set_markup(f"<big><b>{text}</b></big>")
        else:
            self.setup()
            page = self.page_name()

        GLib.idle_add(self.set_initial_page, page)

    def disable_retry_on_clicked(self, widget):
        self.disabled_retry_button.set_sensitive(False)
        self.on_module_selected()

    def set_initial_page(self, page):
        self.sidePage.stack.set_visible_child_name(page)

    def reset(self):
        self.bolt_manager = None
        for obj_path in list(self.bolt_devices.keys()):
            self.bolt_device_removed(obj_path, False)

    def setup(self):
        if not self.bolt_manager:
            self.bolt_manager = BoltManagerProxy()
            self.bolt_manager.on_device_added = self.bolt_device_added
            self.bolt_manager.on_device_removed = self.bolt_device_removed
        for obj_path in self.bolt_manager.list_devices():
            device = BoltDeviceProxy(obj_path)                
            # Skip the host device
            if device.props.Type == "host":
                continue
            # Add the device
            self.bolt_device_added(obj_path, False)

    def bolt_device_added(self, obj_path, change_page=True):
        if obj_path not in self.bolt_devices:
            # Build the section
            device = BoltDeviceProxy(obj_path)
            section = BoltSection(self.bolt_manager, device)
            section.show_all()
            # Add to the page
            self.bolt_devices[obj_path] = section
            page = self.sidePage.stack.get_child_by_name("settings")
            page.pack_start(section, False, False, 0)
        if change_page:
            self.sidePage.stack.set_visible_child_name(self.page_name())

    def bolt_device_removed(self, obj_path, change_page=True):
        if obj_path in self.bolt_devices:
            section = self.bolt_devices[obj_path]
            section.destroy()
            del self.bolt_devices[obj_path]
        if change_page:
            self.sidePage.stack.set_visible_child_name(self.page_name())

    def page_name(self):
        return "settings" if len(self.bolt_devices) > 0 else "empty"

    def test_daemon_alive(self):
        try:
            # Ping bolt to see if its available and running
            Gio.DBusConnection.call_sync(
                Gio.bus_get_sync(Gio.BusType.SYSTEM),
                BOLT_BUS_NAME,
                BOLT_OBJECT_PATH,
                "org.freedesktop.DBus.Peer",
                "Ping",
                None,
                None,
                0,
                -1,
                None)
            return True
        except GLib.Error as e:
            # Bolt isn't installed or service is disabled
            pass
        return False



#!/usr/bin/python3

from gi.repository import Gtk
import pageutils

class ModulePage(pageutils.BaseListView):
    def __init__(self, parent):
        store = Gtk.ListStore(int, str, str, str, str, str)
        pageutils.BaseListView.__init__(self, store)

        self.parent = parent
        self.adjust = self.get_vadjustment()

        column = self.create_text_column(0, "ID")
        column.set_cell_data_func(self.renderer_text, self.cell_data_func_id)
        self.create_text_column(1, "Name")
        self.create_text_column(2, "Type")
        self.create_text_column(3, "Value")
        self.tree_view.set_tooltip_column(4)

        self.tree_view.connect("row-activated", self.on_row_activated)

        self.parent.lg_proxy.connect("signal::result-update", self.get_updates)
        self.parent.lg_proxy.connect("signal::inspector-done", self.on_inspector_done)
        self.parent.lg_proxy.connect("status-changed", self.on_status_change)

        self._changed = False
        self.tree_view.connect("size-allocate", self.scroll_to_bottom)

    def scroll_to_bottom(self, widget, data):
        if self._changed:
            self.adjust.set_value(self.adjust.get_upper() - self.adjust.get_page_size())
            self._changed = False

    def cell_data_func_id(self, column, cell, model, tree_iter, data=None):
        cell.set_property("text", "r(%d)" %  model.get_value(tree_iter, 0))

    def on_row_activated(self, treeview, path, view_column):
        tree_iter = self.store.get_iter(path)
        result_id = self.store.get_value(tree_iter, 0)
        name = self.store.get_value(tree_iter, 1)
        obj_type = self.store.get_value(tree_iter, 2)
        value = self.store.get_value(tree_iter, 5)

        self.parent.pages["inspect"].inspect_element("r(%d)" % result_id, obj_type, name, value)

    def on_status_change(self, proxy, online):
        if online:
            self.get_updates()

    def get_updates(self, proxy=None):
        self.store.clear()
        success, data = self.parent.lg_proxy.GetResults()
        if success:
            try:
                for item in data:
                    self.store.append([int(item["index"]),
                                       item["command"],
                                       item["type"],
                                       pageutils.shorten_value(item["object"]),
                                       item["tooltip"],
                                       item["object"]])
                self._changed = True
            except Exception as exc:
                print(exc)

    def on_inspector_done(self, proxy=None):
        self.parent.show()
        self.parent.activate_page("results")
        self.get_updates()



#!/usr/bin/python3

from gi.repository import GLib, Gio
import sys
import signal
import shlex
import argparse
import gettext
from subprocess import Popen, DEVNULL
from enum import IntEnum

signal.signal(signal.SIGINT, signal.SIG_DFL)
gettext.install("cinnamon", "/usr/share/locale")

# DBus interface constants
SS_SERVICE = "org.cinnamon.ScreenSaver"
SS_PATH = "/org/cinnamon/ScreenSaver"
SS_INTERFACE = "org.cinnamon.ScreenSaver"

class Action(IntEnum):
    EXIT = 1
    QUERY = 2
    TIME = 3
    LOCK = 4
    ACTIVATE = 5
    DEACTIVATE = 6
    VERSION = 7

class ScreensaverCommand:
    """
    Standalone executable for controlling the screensaver via DBus.
    Supports both internal (Main.screenShield) and external (cinnamon-screensaver) modes.
    """
    def __init__(self, mainloop):
        self.mainloop = mainloop
        self.proxy = None

        parser = argparse.ArgumentParser(description='Cinnamon Screensaver Command')
        parser.add_argument('--exit', '-e', dest="action_id", action='store_const', const=Action.EXIT,
                            help=_('Causes the screensaver to exit gracefully'))
        parser.add_argument('--query', '-q', dest="action_id", action='store_const', const=Action.QUERY,
                            help=_('Query the state of the screensaver'))
        parser.add_argument('--time', '-t', dest="action_id", action='store_const', const=Action.TIME,
                            help=_('Query the length of time the screensaver has been active'))
        parser.add_argument('--lock', '-l', dest="action_id", action='store_const', const=Action.LOCK,
                            help=_('Tells the running screensaver process to lock the screen immediately'))
        parser.add_argument('--activate', '-a', dest="action_id", action='store_const', const=Action.ACTIVATE,
                            help=_('Turn the screensaver on (blank the screen)'))
        parser.add_argument('--deactivate', '-d', dest="action_id", action='store_const', const=Action.DEACTIVATE,
                            help=_('If the screensaver is active then deactivate it (un-blank the screen)'))
        parser.add_argument('--version', '-V', dest="action_id", action='store_const', const=Action.VERSION,
                            help=_('Version of this application'))
        parser.add_argument('--away-message', '-m', dest="message", action='store', default="",
                            help=_('Message to be displayed in lock screen'))
        args = parser.parse_args()

        if not args.action_id:
            parser.print_help()
            quit()

        if args.action_id == Action.VERSION:
            # Get version from cinnamon
            try:
                version_proxy = Gio.DBusProxy.new_for_bus_sync(
                    Gio.BusType.SESSION,
                    Gio.DBusProxyFlags.NONE,
                    None,
                    'org.Cinnamon',
                    '/org/Cinnamon',
                    'org.Cinnamon',
                    None
                )
                version = version_proxy.get_cached_property('CinnamonVersion')
                if version:
                    print("cinnamon-screensaver-command (Cinnamon %s)" % version.unpack())
                else:
                    print("cinnamon-screensaver-command (Cinnamon version unknown)")
            except:
                print("cinnamon-screensaver-command")
            quit()

        self.action_id = args.action_id
        self.message = args.message

        ss_settings = Gio.Settings.new("org.cinnamon.desktop.screensaver")
        custom_saver = ss_settings.get_string("custom-screensaver-command").strip()
        internal_enabled = Gio.Settings.new("org.cinnamon").get_boolean("internal-screensaver-enabled")

        # Cinnamon dismisses its own modals inside the internal shield path.
        # Only call it here when the lock is being handled by a custom command
        # or cinnamon-screensaver.
        if self.action_id in (Action.LOCK, Action.ACTIVATE) and (custom_saver or not internal_enabled):
            self._dismiss_cinnamon_modals()

        if custom_saver:
            self._handle_custom_saver(custom_saver)
            quit()

        # Create DBus proxy
        Gio.DBusProxy.new_for_bus(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            SS_SERVICE,
            SS_PATH,
            SS_INTERFACE,
            None,
            self._on_proxy_ready
        )

    def _dismiss_cinnamon_modals(self):
        return
        try:
            bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            bus.call_sync("org.Cinnamon",
                          "/org/Cinnamon",
                          "org.Cinnamon",
                          "DismissInternalModals",
                          None,
                          None,
                          Gio.DBusCallFlags.NONE,
                          2000,
                          None)
        except GLib.Error as e:
            print("Error dismissing Cinnamon modals: %s" % e.message)

    def _handle_custom_saver(self, custom_saver):
        if self.action_id in (Action.LOCK, Action.ACTIVATE):
            try:
                Popen(shlex.split(custom_saver), stdin=DEVNULL)
            except OSError as e:
                print("Error %d running %s: %s" % (e.errno, custom_saver, e.strerror))
        else:
            print("Action not supported with custom screensaver.")

    def _on_proxy_ready(self, source, result):
        try:
            self.proxy = Gio.DBusProxy.new_for_bus_finish(result)
            self.perform_action()
        except GLib.Error as e:
            print("Can't connect to screensaver: %s" % e.message)
            self.mainloop.quit()

    def perform_action(self):
        try:
            if self.action_id == Action.EXIT:
                self.proxy.call_sync(
                    'Quit',
                    None,
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                )

            elif self.action_id == Action.QUERY:
                result = self.proxy.call_sync(
                    'GetActive',
                    None,
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                )
                if result:
                    is_active = result.unpack()[0]
                    if is_active:
                        print(_("The screensaver is active\n"))
                    else:
                        print(_("The screensaver is inactive\n"))

            elif self.action_id == Action.TIME:
                result = self.proxy.call_sync(
                    'GetActiveTime',
                    None,
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                )
                if result:
                    time = result.unpack()[0]
                    if time == 0:
                        print(_("The screensaver is not currently active.\n"))
                    else:
                        print(gettext.ngettext(
                            "The screensaver has been active for %d second.\n",
                            "The screensaver has been active for %d seconds.\n",
                            time
                        ) % time)

            elif self.action_id == Action.LOCK:
                self.proxy.call_sync(
                    'Lock',
                    GLib.Variant('(s)', (self.message,)),
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                )

            elif self.action_id == Action.ACTIVATE:
                self.proxy.call_sync(
                    'SetActive',
                    GLib.Variant('(b)', (True,)),
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                )

            elif self.action_id == Action.DEACTIVATE:
                self.proxy.call_sync(
                    'SetActive',
                    GLib.Variant('(b)', (False,)),
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                )

        except GLib.Error as e:
            print("Error executing command: %s" % e.message)

        self.mainloop.quit()

if __name__ == "__main__":
    ml = GLib.MainLoop.new(None, True)
    main = ScreensaverCommand(ml)
    ml.run()



#!/usr/bin/python3

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, GObject
from PIL import Image

class EyeDropper(Gtk.HBox):
    __gsignals__ = {
        'color-picked': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_STRING,))
    }

    def __init__(self):
        Gtk.HBox.__init__ (self)

        self.button = Gtk.Button("")
        self.button.set_tooltip_text(_("Click the eyedropper, then click a color anywhere on your screen to select that color"))
        self.button.set_image(Gtk.Image().new_from_stock(Gtk.STOCK_COLOR_PICKER, Gtk.IconSize.BUTTON))
        self.button.get_property('image').show()
        self.button.set_events(Gdk.EventMask.POINTER_MOTION_MASK | Gdk.EventMask.POINTER_MOTION_HINT_MASK)

        self.pack_start(self.button, False, False, 2)

        self.bp_handler = None
        self.br_handler = None
        self.kp_handler = None

        self.button.connect("clicked", self.on_button_clicked)

    def on_button_clicked(self, widget):
        screen = widget.get_screen()
        self.time = Gtk.get_current_event_time()
        self.device = Gtk.get_current_event_device()

        self.grab_widget = Gtk.Window(Gtk.WindowType.POPUP)
        self.grab_widget.set_screen(screen)
        self.grab_widget.resize(1, 1)
        self.grab_widget.move(-100, -100)
        self.grab_widget.show()

        self.grab_widget.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.POINTER_MOTION_MASK)
        toplevel = widget.get_toplevel()

        if isinstance(toplevel, Gtk.Window):
            if toplevel.has_group():
                toplevel.add_window(grab_widget)

        window = self.grab_widget.get_window()

        picker_cursor = Gdk.Cursor(screen.get_display(), Gdk.CursorType.CROSSHAIR)

        grab_status = self.device.grab(window, Gdk.GrabOwnership.APPLICATION, False,
                                       Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.POINTER_MOTION_MASK,
                                       picker_cursor, self.time)

        if grab_status != Gdk.GrabStatus.SUCCESS:
            return

        Gtk.device_grab_add(self.grab_widget, self.device, True)

        self.bp_handler = self.grab_widget.connect("button-press-event", self.mouse_press)
        self.kp_handler = self.grab_widget.connect("key-press-event", self.key_press)

    def mouse_press(self, widget, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            self.br_handler = widget.connect("button-release-event", self.mouse_release)
            return True
        return False

    def key_press(self, widget, event):
        screen, x_root, y_root = self.device.get_position()
        if event.keyval == Gdk.KEY_Escape:
            self.ungrab(self.device)
            return True
        elif event.keyval in (Gdk.KEY_space, Gdk.KEY_Return, Gdk.KEY_ISO_Enter, Gdk.KEY_KP_Enter, Gdk.KEY_KP_Space):
            self.grab_color_at_pointer(event, screen, x_root, y_root)
            return True
        return False

    def mouse_release(self, widget, event):
        screen, x, y = self.device.get_position()
        if event.button != 1:
            return False
        self.grab_color_at_pointer(event, screen, event.x_root, event.y_root)
        return True

    def grab_color_at_pointer(self, event, screen, x_root, y_root):
        device = self.device
        window = screen.get_root_window()
        pixbuf = Gdk.pixbuf_get_from_window(window, x_root, y_root, 1, 1)
        image = pixbuf2Image(pixbuf)

        r, g, b = image.getpixel((0, 0))

        color = Gdk.RGBA()
        color.red = r / 255.0
        color.green = g / 255.0
        color.blue = b / 255.0
        self.emit('color-picked', color.to_string())
        self.ungrab(device)

    def ungrab(self, device):
        device.ungrab(self.time)
        Gtk.device_grab_remove(self.grab_widget, device)
        self.grab_widget.handler_disconnect(self.bp_handler)
        self.grab_widget.handler_disconnect(self.br_handler)
        self.grab_widget.handler_disconnect(self.kp_handler)

def pixbuf2Image(pb):
    width,height = pb.get_width(),pb.get_height()
    return Image.fromstring("RGB",(width,height),pb.get_pixels() )



#!/usr/bin/python3

import os
import subprocess

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk, GObject, GLib

from xapp.SettingsWidgets import SettingsWidget, SettingsLabel
from xapp.GSettingsWidgets import PXGSettingsBackend
from bin.ChooserButtonWidgets import DateChooserButton, TimeChooserButton
from bin.KeybindingWidgets import ButtonKeybinding
from bin import util

from bin import KeybindingTable

settings_objects = {}

CAN_BACKEND = ["SoundFileChooser", "DateChooser", "TimeChooser", "Keybinding"]

class BinFileMonitor(GObject.GObject):
    __gsignals__ = {
        'changed': (GObject.SignalFlags.RUN_LAST, None, ()),
    }
    def __init__(self):
        super(BinFileMonitor, self).__init__()

        self.changed_id = 0

        env = GLib.getenv("PATH")

        if env is None:
            env = "/bin:/usr/bin:."

        self.paths = env.split(":")

        self.monitors = []

        for path in self.paths:
            file = Gio.File.new_for_path(path)
            try:
                mon = file.monitor_directory(Gio.FileMonitorFlags.SEND_MOVED, None)
                mon.connect("changed", self.queue_emit_changed)
                self.monitors.append(mon)
            except GLib.Error as e:
                pass

    def _emit_changed(self):
        self.emit("changed")
        self.changed_id = 0
        return False

    def queue_emit_changed(self, file, other, event_type, data=None):
        if self.changed_id > 0:
            GLib.source_remove(self.changed_id)
            self.changed_id = 0

        self.changed_id = GLib.idle_add(self._emit_changed)

file_monitor = None

def get_file_monitor():
    global file_monitor

    if file_monitor is None:
        file_monitor = BinFileMonitor()

    return file_monitor

class DependencyCheckInstallButton(Gtk.Box):
    def __init__(self, checking_text, install_button_text, binfiles, final_widget=None, satisfied_cb=None):
        super(DependencyCheckInstallButton, self).__init__(orientation=Gtk.Orientation.HORIZONTAL)

        self.binfiles = binfiles
        self.satisfied_cb = satisfied_cb

        self.checking_text = checking_text
        self.install_button_text = install_button_text

        self.stack = Gtk.Stack()
        self.pack_start(self.stack, False, False, 0)

        self.progress_bar = Gtk.ProgressBar()
        self.stack.add_named(self.progress_bar, "progress")

        self.progress_bar.set_show_text(True)
        self.progress_bar.set_text(self.checking_text)

        self.install_warning = Gtk.Label(label=install_button_text, margin=5)
        frame = Gtk.Frame()
        frame.add(self.install_warning)
        frame.set_shadow_type(Gtk.ShadowType.OUT)
        frame.show_all()
        self.stack.add_named(frame, "install")

        if final_widget:
            self.stack.add_named(final_widget, "final")
        else:
            self.stack.add_named(Gtk.Alignment(), "final")

        self.stack.set_visible_child_name("progress")
        self.progress_source_id = 0

        self.file_listener = get_file_monitor()
        self.file_listener_id = self.file_listener.connect("changed", self.on_file_listener_ping)

        self.connect("destroy", self.on_destroy)

        GLib.idle_add(self.check)

    def check(self):
        self.start_pulse()

        success = True

        for program in self.binfiles:
            if not GLib.find_program_in_path(program):
                success = False
                break

        GLib.idle_add(self.on_check_complete, success)

        return False

    def pulse_progress(self):
        self.progress_bar.pulse()
        return True

    def start_pulse(self):
        self.cancel_pulse()
        self.progress_source_id = GLib.timeout_add(200, self.pulse_progress)

    def cancel_pulse(self):
        if self.progress_source_id > 0:
            GLib.source_remove(self.progress_source_id)
            self.progress_source_id = 0

    def on_check_complete(self, result, data=None):
        self.cancel_pulse()
        if result:
            self.stack.set_visible_child_name("final")
            if self.satisfied_cb:
                self.satisfied_cb()
        else:
            self.stack.set_visible_child_name("install")

    def on_file_listener_ping(self, monitor, data=None):
        self.stack.set_visible_child_name("progress")
        self.progress_bar.set_text(self.checking_text)
        self.check()

    def on_destroy(self, widget):
        self.file_listener.disconnect(self.file_listener_id)
        self.file_listener_id = 0

class GSettingsDependencySwitch(SettingsWidget):
    def __init__(self, label, schema=None, key=None, dep_key=None, binfiles=None, packages=None):
        super(GSettingsDependencySwitch, self).__init__(dep_key=dep_key)

        self.binfiles = binfiles
        self.packages = packages

        self.content_widget = Gtk.Alignment()
        self.label = Gtk.Label(label)
        self.pack_start(self.label, False, False, 0)
        self.pack_end(self.content_widget, False, False, 0)

        self.switch = Gtk.Switch()
        self.switch.set_halign(Gtk.Align.END)
        self.switch.set_valign(Gtk.Align.CENTER)

        pkg_string = ""
        for pkg in packages:
            if pkg_string != "":
                pkg_string += ", "
            pkg_string += pkg

        self.dep_button = DependencyCheckInstallButton(_("Checking dependencies"),
                                                       _("Please install: %s") % pkg_string,
                                                       binfiles,
                                                       self.switch)
        self.content_widget.add(self.dep_button)

        if schema:
            self.settings = self.get_settings(schema)
            self.settings.bind(key, self.switch, "active", Gio.SettingsBindFlags.DEFAULT)

class SidePage(object):
    def __init__(self, name, icon, keywords, content_box = None, size = None, is_c_mod = False, is_standalone = False, exec_name = None, module=None):
        self.name = name
        self.icon = icon
        self.content_box = content_box
        self.widgets = []
        self.is_c_mod = is_c_mod
        self.is_standalone = is_standalone
        self.exec_name = exec_name
        self.module = module # Optionally set by the module so we can call on_module_selected() on it when we show it.
        self.keywords = keywords
        self.size = size
        self.topWindow = None
        self.builder = None
        self.stack = None
        if self.module is not None:
            self.module.loaded = False

    def add_widget(self, widget):
        self.widgets.append(widget)

    def build(self):
        # Clear all the widgets from the content box
        widgets = self.content_box.get_children()
        for widget in widgets:
            self.content_box.remove(widget)

        if self.module is not None:
            self.module.on_module_selected()
            self.module.loaded = True

        if self.is_standalone:
            subprocess.Popen(self.exec_name.split())
            return

        # Add our own widgets
        for widget in self.widgets:
            if hasattr(widget, 'expand'):
                self.content_box.pack_start(widget, True, True, 2)
            else:
                self.content_box.pack_start(widget, False, False, 2)

        # C modules are sort of messy - they check the desktop type
        # (for Unity or GNOME) and show/hide UI items depending on
        # the result - so we cannot just show_all on the widget, it will
        # mess up these modifications - so for these, we just show the
        # top-level widget
        if not self.is_c_mod:
            self.content_box.show_all()
            try:
                self.check_third_arg()
            except:
                pass
            return

        self.content_box.show()
        for child in self.content_box:
            child.show()

            # C modules can have non-C parts. C parts are all named c_box
            if child.get_name() != "c_box":
                pass

            c_widgets = child.get_children()
            if not c_widgets:
                c_widget = self.content_box.c_manager.get_c_widget(self.exec_name)
                if c_widget is not None:
                    child.pack_start(c_widget, False, False, 2)
                    c_widget.show()
            else:
                for c_widget in c_widgets:
                    c_widget.show()

            def recursively_iterate(parent):
                if self.stack:
                    return
                for child in parent:
                    if isinstance(child, Gtk.Stack):
                        self.stack = child
                        break
                    elif isinstance(child, Gtk.Container):
                        recursively_iterate(child)

            # Look for a stack recursively
            recursively_iterate(child)

class CCModule:
    def __init__(self, label, mod_id, icon, category, keywords, content_box):
        sidePage = SidePage(label, icon, keywords, content_box, size=-1, is_c_mod=True, is_standalone=False, exec_name=mod_id, module=None)
        self.sidePage = sidePage
        self.name = mod_id
        self.category = category

    def process (self, c_manager):
        if c_manager.lookup_c_module(self.name):
            c_box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 2)
            c_box.set_vexpand(False)
            c_box.set_name("c_box")
            self.sidePage.add_widget(c_box)
            return True
        else:
            return False

class SAModule:
    def __init__(self, label, mod_id, icon, category, keywords, content_box):
        sidePage = SidePage(label, icon, keywords, content_box, False, False, True, mod_id)
        self.sidePage = sidePage
        self.name = mod_id
        self.category = category

    def process (self):
        name = self.name.replace("pkexec ", "")
        name = name.split()[0]

        return GLib.find_program_in_path(name) is not None

def walk_directories(dirs, filter_func, return_directories=False):
    # If return_directories is False: returns a list of valid subdir names
    # Else: returns a list of valid tuples (subdir-names, parent-directory)
    valid = []
    try:
        for thdir in dirs:
            if os.path.isdir(thdir):
                for t in os.listdir(thdir):
                    if filter_func(os.path.join(thdir, t)):
                        if return_directories:
                            valid.append([t, thdir])
                        else:
                            valid.append(t)
    except:
        pass
        #logging.critical("Error parsing directories", exc_info=True)
    return valid

class LabelRow(SettingsWidget):
    def __init__(self, text=None, tooltip=None):
        super(LabelRow, self).__init__()

        self.label = SettingsLabel()
        self.label.set_hexpand(True)
        self.pack_start(self.label, False, False, 0)
        self.label.set_markup(text)
        self.set_tooltip_text(tooltip)

class SoundFileChooser(SettingsWidget):
    bind_dir = None

    def __init__(self, label, event_sounds=True, size_group=None, dep_key=None, tooltip=""):
        super(SoundFileChooser, self).__init__(dep_key=dep_key)

        self.event_sounds = event_sounds

        self.label = SettingsLabel(label)
        self.content_widget = Gtk.Box()

        c = self.content_widget.get_style_context()
        c.add_class(Gtk.STYLE_CLASS_LINKED)

        self.file_picker_button = Gtk.Button()
        self.file_picker_button.connect("clicked", self.on_picker_clicked)

        button_content = Gtk.Box(spacing=5)
        self.file_picker_button.add(button_content)

        self.button_label = Gtk.Label()
        button_content.pack_start(Gtk.Image(icon_name="sound"), False, False, 0)
        button_content.pack_start(self.button_label, False, False, 0)

        self.content_widget.pack_start(self.file_picker_button, True, True, 0)

        self.pack_start(self.label, False, False, 0)
        self.pack_end(self.content_widget, False, False, 0)

        self.play_button = Gtk.Button()
        self.play_button.set_image(Gtk.Image.new_from_icon_name("xsi-media-playback-start-symbolic", Gtk.IconSize.BUTTON))
        self.play_button.connect("clicked", self.on_play_clicked)
        self.content_widget.pack_start(self.play_button, False, False, 0)

        self.set_tooltip_text(tooltip)

        if size_group:
            self.add_to_size_group(size_group)

    def on_play_clicked(self, widget):
        util.play_sound_file(self.get_value())

    def on_picker_clicked(self, widget):
        dialog = Gtk.FileChooserDialog(title=self.label.get_text(),
                                       action=Gtk.FileChooserAction.OPEN,
                                       transient_for=self.get_toplevel(),
                                       buttons=(_("_Cancel"), Gtk.ResponseType.CANCEL,
                                                _("_Open"), Gtk.ResponseType.ACCEPT))

        if os.path.exists(self.get_value()):
            dialog.set_filename(self.get_value())
        else:
            dialog.set_current_folder('/usr/share/sounds')

        sound_filter = Gtk.FileFilter()
        if self.event_sounds:
            sound_filter.add_mime_type("audio/x-wav")
            sound_filter.add_mime_type("audio/x-vorbis+ogg")
        else:
            sound_filter.add_mime_type("audio/*")
        sound_filter.set_name(_("Sound files"))
        dialog.add_filter(sound_filter)

        if dialog.run() == Gtk.ResponseType.ACCEPT:
            name = dialog.get_filename()
            self.set_value(name)
            self.update_button_label(name)

        dialog.destroy()

    def update_button_label(self, absolute_path):
        if absolute_path != "":
            f = Gio.File.new_for_path(absolute_path)
            self.button_label.set_label(f.get_basename())

    def on_setting_changed(self, *args):
        self.update_button_label(self.get_value())

    def connect_widget_handlers(self, *args):
        pass

class DateChooser(SettingsWidget):
    bind_dir = None

    def __init__(self, label, size_group=None, dep_key=None, tooltip=""):
        super(DateChooser, self).__init__(dep_key=dep_key)

        self.label = SettingsLabel(label)

        self.content_widget = DateChooserButton()

        self.pack_start(self.label, False, False, 0)
        self.pack_end(self.content_widget, False, False, 0)

        self.set_tooltip_text(tooltip)

        if size_group:
            self.add_to_size_group(size_group)

    def on_date_changed(self, *args):
        date = self.content_widget.get_date()
        self.set_value({"y": date.year, "m": date.month, "d": date.day})

    def on_setting_changed(self, *args):
        date = self.get_value()
        self.content_widget.set_date((date["y"], date["m"], date["d"]))

    def connect_widget_handlers(self, *args):
        self.content_widget.connect("date-changed", self.on_date_changed)

class TimeChooser(SettingsWidget):
    bind_dir = None

    def __init__(self, label, size_group=None, dep_key=None, tooltip=""):
        super(TimeChooser, self).__init__(dep_key=dep_key)

        self.label = SettingsLabel(label)

        self.content_widget = TimeChooserButton()

        self.pack_start(self.label, False, False, 0)
        self.pack_end(self.content_widget, False, False, 0)

        self.set_tooltip_text(tooltip)

        if size_group:
            self.add_to_size_group(size_group)

    def on_time_changed(self, *args):
        time = self.content_widget.get_time()
        self.set_value({"h": time.hour, "m": time.minute, "s": time.second})

    def on_setting_changed(self, *args):
        time = self.get_value()
        self.content_widget.set_time((time["h"], time["m"], time["s"]))

    def connect_widget_handlers(self, *args):
        self.content_widget.connect("time-changed", self.on_time_changed)

class Keybinding(SettingsWidget):
    bind_dir = None

    def __init__(self, label, num_bind=2, size_group=None, dep_key=None, tooltip=""):
        super(Keybinding, self).__init__(dep_key=dep_key)

        self.num_bind = num_bind
        self.kb_table = KeybindingTable.get_default()
        self.kb_label = label

        if self.backend == "gsettings":
            self.keybinding = self.kb_table.lookup_gsettings_keybinding(self.settings.props.schema_id, self.key)
        else:
            self.keybinding = self.kb_table.lookup_json_keybinding(self.settings.uuid, self.settings.instance_id, self.key)

        self.keybinding.connect("changed", self.on_kb_table_entry_changed)

        self.label = SettingsLabel(label)

        self.buttons = []
        self.teach_button = None

        self.content_widget = Gtk.Frame(shadow_type=Gtk.ShadowType.IN)
        self.content_widget.set_valign(Gtk.Align.CENTER)
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.content_widget.add(box)

        self.pack_start(self.label, False, False, 0)
        self.pack_end(self.content_widget, False, False, 0)

        for x in range(self.num_bind):
            if x != 0:
                box.add(Gtk.Separator(orientation=Gtk.Orientation.VERTICAL))
            kb = ButtonKeybinding(position=x)
            kb.set_size_request(150, -1)
            kb.set_accel_string(self.keybinding.entries[x] if self.keybinding else "")
            kb.connect("accel-edited", self.on_kb_changed)
            kb.connect("accel-cleared", self.on_kb_cleared)
            box.pack_start(kb, False, False, 0)
            self.buttons.append(kb)

        self.event_id = None
        self.teaching = False

        self.set_tooltip_text(tooltip)

        if size_group:
            self.add_to_size_group(size_group)

    def on_kb_changed(self, button, accel_string, accel_label):
        if self.keybinding:
            if not self.kb_table.maybe_update_binding(self.keybinding, accel_string, accel_label, button.position):
                self.on_kb_table_entry_changed(self.keybinding)

    def on_kb_cleared(self, button):
        if self.keybinding:
            self.kb_table.clear_binding(self.keybinding, button.position)

    def on_kb_table_entry_changed(self, keybinding):
        for x in range(0, self.num_bind):
            try:
                self.buttons[x].set_accel_string(keybinding.entries[x])
            except IndexError:
                self.buttons[x].set_accel_string("")

    def on_setting_changed(self, *args):
        pass

    def connect_widget_handlers(self, *args):
        pass

def g_settings_factory(subclass):
    class NewClass(globals()[subclass], PXGSettingsBackend):
        def __init__(self, label, schema, key, *args, **kwargs):
            self.backend = "gsettings"
            self.key = key
            if schema not in settings_objects:
                settings_objects[schema] = Gio.Settings(schema_id=schema)
            self.settings = settings_objects[schema]

            if "map_get" in kwargs:
                self.map_get = kwargs["map_get"]
                del kwargs["map_get"]
            if "map_set" in kwargs:
                self.map_set = kwargs["map_set"]
                del kwargs["map_set"]

            super(NewClass, self).__init__(label, *args, **kwargs)
            self.bind_settings()
    return NewClass

for widget in CAN_BACKEND:
    globals()["GSettings"+widget] = g_settings_factory(widget)



#!/usr/bin/python3

import gi
gi.require_version('Notify', '0.7')
from gi.repository import Gio, Notify, Gtk, Pango
import re

from bin.SettingsWidgets import SidePage
from xapp.GSettingsWidgets import *

content = """
Lorem ipsum dolor sit amet, consectetur adipiscing elit. \
Suspendisse eleifend, lacus ut tempor vehicula, lorem tortor \
suscipit libero, sit amet congue odio libero vitae lacus. \
Sed est nibh, lacinia ac magna non, blandit aliquet est. \
Mauris volutpat est vel lacinia faucibus. Pellentesque \
pulvinar eros at dolor pretium, eget hendrerit leo rhoncus. \
Sed nisl leo, posuere eget risus vel, euismod egestas metus. \
Praesent interdum, dui sit amet convallis rutrum, velit nunc \
sollicitudin erat, ac viverra leo eros in nulla. Morbi feugiat \
feugiat est. Nam non libero dolor. Duis egestas sodales massa \
sit amet lobortis. Donec sit amet nisi turpis. Morbi aliquet \
aliquam ullamcorper.
"""

NOTIFICATION_DISPLAY_SCREENS = [
    ("primary-screen", _("Primary monitor")),
    ("active-screen", _("Active monitor")),
    ("fixed-screen", _("The monitor specified below"))
]


class Module:
    name = "notifications"
    comment = _("Notification preferences")
    category = "prefs"

    def __init__(self, content_box):
        keywords = _("notifications")
        sidePage = SidePage(_("Notifications"), "cs-notifications", keywords, content_box, module=self)
        self.sidePage = sidePage

    def on_module_selected(self):
        if self.loaded:
            return

        print("Loading Notifications module")

        Notify.init("cinnamon-settings-notifications-test")

        page = SettingsPage()
        self.sidePage.add_widget(page)

        settings = page.add_section()

        switch = GSettingsSwitch(_("Enable notifications"), "org.cinnamon.desktop.notifications", "display-notifications")
        settings.add_row(switch)

        button = Button(_("Application notifications"), self.open_app_settings)
        settings.add_reveal_row(button, "org.cinnamon.desktop.notifications", "display-notifications")

        settings = page.add_reveal_section(_("Notification settings"), "org.cinnamon.desktop.notifications", "display-notifications")

        switch = GSettingsSwitch(_("Remove notifications after their timeout is reached"), "org.cinnamon.desktop.notifications", "remove-old")
        settings.add_reveal_row(switch, "org.cinnamon.desktop.notifications", "display-notifications")

        switch = GSettingsSwitch(_("Show notifications on the bottom side of the screen"), "org.cinnamon.desktop.notifications", "bottom-notifications")
        settings.add_reveal_row(switch, "org.cinnamon.desktop.notifications", "display-notifications")

        combo = GSettingsComboBox(_("Monitor to use for displaying notifications"), "org.cinnamon.desktop.notifications", "notification-screen-display", NOTIFICATION_DISPLAY_SCREENS)
        settings.add_reveal_row(combo, "org.cinnamon.desktop.notifications", "display-notifications")

        spin = GSettingsSpinButton(_("Monitor"), "org.cinnamon.desktop.notifications", "notification-fixed-screen", None, 1, 13, 1)
        settings.add_reveal_row(spin)
        spin.revealer.settings = Gio.Settings("org.cinnamon.desktop.notifications")
        spin.revealer.settings.bind_with_mapping("notification-screen-display", spin.revealer, "reveal-child", Gio.SettingsBindFlags.GET, lambda option: option == "fixed-screen", None)

        switch = GSettingsSwitch(_("Display notifications over fullscreen windows"), "org.cinnamon.desktop.notifications", "fullscreen-notifications")
        settings.add_reveal_row(switch, "org.cinnamon.desktop.notifications", "display-notifications")

        spin = GSettingsSpinButton(_("Notification duration"), "org.cinnamon.desktop.notifications", "notification-duration", _("seconds"), 1, 60, 1, 1)
        settings.add_reveal_row(spin, "org.cinnamon.desktop.notifications", "display-notifications")

        button = Button(_("Display a test notification"), self.send_test)
        settings.add_reveal_row(button, "org.cinnamon.desktop.notifications", "display-notifications")

        settings = page.add_section(_("Media keys OSD"))

        switch = GSettingsSwitch(_("Show media keys OSD"), "org.cinnamon", "show-media-keys-osd")
        settings.add_row(switch)

    def send_test(self, widget):
        n = Notify.Notification.new(_("This is a test notification"), content, "dialog-warning")
        n.show()

    def open_app_settings(self, widget):
        win = AppNotificationsWindow(widget.get_toplevel())

PER_APP_SCHEMA = "org.cinnamon.desktop.notifications.application"
PER_APP_BASE_PATH = "/org/cinnamon/desktop/notifications/application/"

class AppNotificationRow(Gtk.ListBoxRow):
    def __init__(self, app_info, parent_settings):
        super().__init__()
        self.parent_settings = parent_settings
        self.set_activatable(True)
        self.set_selectable(False)
        self.set_can_focus(True)

        self.app_name = app_info.get_name().lower()

        # Sanitise app ID for GSettings path (this should remain the same as in ui/messageTray.js)
        # 1. Convert to lower case.
        # 2. Replace any one or more consecutive characters that is not a lowercase letter or a digit with a hyphen.
        # 3. Trim any leading or trailing hyphens.
        app_id = app_info.get_id().lower().replace(".desktop", "")
        self.settings_id = re.sub(r'[^a-z0-9]+', '-', app_id).strip('-')
        path = f"{PER_APP_BASE_PATH}{self.settings_id}/"

        self.settings = Gio.Settings.new_with_path(PER_APP_SCHEMA, path)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        hbox.set_margin_start(8)
        hbox.set_margin_end(8)
        hbox.set_margin_top(4)
        hbox.set_margin_bottom(4)

        # Icon
        gicon = app_info.get_icon()
        if not gicon:
            gicon = Gio.ThemedIcon.new("application-x-executable")
        icon = Gtk.Image.new_from_gicon(gicon, Gtk.IconSize.DND)
        icon.set_pixel_size(32)
        hbox.pack_start(icon, False, False, 0)

        # Labels
        name_label = Gtk.Label(label=app_info.get_name(), xalign=0)
        name_label.set_ellipsize(Pango.EllipsizeMode.END)
        hbox.pack_start(name_label, True, True, 0)

        # Switch
        self.switch = Gtk.Switch()
        self.switch.set_active(self.settings.get_boolean("enabled"))
        self.settings.bind("enabled", self.switch, "active", Gio.SettingsBindFlags.DEFAULT)
        self.settings.connect("changed::enabled", self.update_index)
        hbox.pack_start(self.switch, False, False, 0)

        self.add(hbox)

    def update_index(self, settings, key):
        current_children = list(self.parent_settings.get_strv("application-children"))

        if self.settings.get_boolean("enabled"):
            # Since 'true' is the default, we can remove the custom setting from dconf
            if self.settings_id in current_children:
                current_children.remove(self.settings_id)
                self.parent_settings.set_strv("application-children", current_children)
                self.settings.reset("enabled")
        else:
            if self.settings_id not in current_children:
                current_children.append(self.settings_id)
                self.parent_settings.set_strv("application-children", current_children)

    def toggle_switch(self):
        self.switch.set_active(not self.switch.get_active())

class AppNotificationsWindow(Gtk.Dialog):
    def __init__(self, parent):
        super().__init__(title=_("Application Notifications"), transient_for=parent)
        self.set_modal(True)
        self.set_destroy_with_parent(True)
        self.set_default_size(430, 480)
        self.set_border_width(10)

        frame = Gtk.Frame()
        frame.set_border_width(6)
        frame.set_shadow_type(Gtk.ShadowType.IN)
        inner_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.search_entry = Gtk.SearchEntry()
        self.search_entry.set_margin_start(16)
        self.search_entry.set_margin_end(16)
        self.search_entry.connect("search-changed", self.on_search_changed)
        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self.listbox = Gtk.ListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.listbox.connect("row-activated", self.on_row_activated)
        self.listbox.set_filter_func(self.filter_func)

        self.parent_settings = Gio.Settings.new("org.cinnamon.desktop.notifications")
        apps = Gio.AppInfo.get_all()
        # Filter for unique apps that are not hidden
        seen_ids = set()
        for app in sorted(apps, key=lambda x: x.get_name()):
            app_id = app.get_id()
            if app.should_show() and app_id not in seen_ids and not app_id.startswith("cinnamon-settings-"):
                row = AppNotificationRow(app, self.parent_settings)
                self.listbox.add(row)
                seen_ids.add(app_id)

        scrolled.add(self.listbox)
        inner_vbox.pack_start(self.search_entry, False, False, 6)
        inner_vbox.pack_start(scrolled, True, True, 0)
        frame.add(inner_vbox)
        content_area = self.get_content_area()
        content_area.pack_start(frame, True, True, 0)

        reset_button = Gtk.Button(label=_("Reset All"))
        reset_button.connect("clicked", self.on_reset_all_clicked)
        self.add_action_widget(reset_button, Gtk.ResponseType.NONE)

        self.show_all()

    def on_row_activated(self, listbox, row):
        row.toggle_switch()

    def filter_func(self, row):
        search_text = self.search_entry.get_text().lower()
        if not search_text:
            return True
        return search_text in row.app_name

    def on_search_changed(self, entry):
        self.listbox.invalidate_filter()

    def on_reset_all_clicked(self, button):
        overridden_apps = self.parent_settings.get_strv("application-children")
        if not overridden_apps:
            return

        for app_id in overridden_apps:
            path = f"{PER_APP_BASE_PATH}{app_id}/"
            app_settings = Gio.Settings.new_with_path(PER_APP_SCHEMA, path)
            app_settings.reset("enabled")

        self.parent_settings.set_strv("application-children", [])



#!/usr/bin/python3

from bin.ExtensionCore import ManageSpicesPage, DownloadSpicesPage
from bin.SettingsWidgets import SidePage
from xapp.SettingsWidgets import SettingsStack
from bin.Spices import Spice_Harvester
from gi.repository import GLib

class Module:
    name = "extensions"
    category = "prefs"
    comment = _("Manage your Cinnamon extensions")

    def __init__(self, content_box):
        self.window = None
        self.sidePage = ExtensionViewSidePage(content_box, self)

    def on_module_selected(self):
        if not self.loaded:
            print("Loading Extensions module")
            self.sidePage.load(self.window)

    def _setParentRef(self, window):
        self.window = window

class ExtensionViewSidePage(SidePage):
    collection_type = "extension"

    def __init__(self, content_box, module):
        self.RemoveString = ""
        keywords = _("extension, addon")

        super(ExtensionViewSidePage, self).__init__(_("Extensions"), "cs-extensions", keywords, content_box, module=module)

    def load(self, window):
        self.window = window

        self.spices = Spice_Harvester(self.collection_type, self.window)

        self.stack = SettingsStack()
        self.add_widget(self.stack)
        self.stack.expand = True

        manage_extensions_page = ManageExtensionsPage(self, self.spices, window)
        self.stack.add_titled(manage_extensions_page, "installed", _("Manage"))

        download_extensions_page = DownloadSpicesPage(self, self.collection_type, self.spices, window)
        self.stack.add_titled(download_extensions_page, "more", _("Download"))

class ManageExtensionsPage(ManageSpicesPage):
    directories = ['/usr/share/cinnamon/extensions', f"{GLib.get_home_dir()}/.local/share/cinnamon/extensions"]
    collection_type = "extension"
    installed_page_title = _("Installed extensions")
    instance_button_text = _("Enable")
    remove_button_text = _("Disable")
    uninstall_button_text = _("Uninstall")
    restore_button_text = _("Disable all")

    def __init__(self, parent, spices, window):
        super(ManageExtensionsPage, self).__init__(parent, self.collection_type, spices, window)



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, GLib

from backends.base import Backend


def _grab_root(x=0, y=0, w=None, h=None):
    display = Gdk.Display.get_default()
    if display is None:
        return None
    root = display.get_default_screen().get_root_window()
    if w is None:
        w = root.get_width()
    if h is None:
        h = root.get_height()
    return Gdk.pixbuf_get_from_window(root, x, y, w, h)


def _deliver(on_done, result):
    def cb():
        on_done(result)
        return GLib.SOURCE_REMOVE
    GLib.idle_add(cb)


class X11Backend(Backend):
    """Fallback backend used when the cinnamon screenshot DBus service is
    unowned. Supports full-screen and arbitrary-rect capture (used for
    per-monitor grabs); window capture and the interactive area selector
    require the cinnamon service. include_pointer is ignored — the X11
    grab path does not overlay the cursor. Results are delivered on the
    main loop's next idle iteration so the call shape matches the async
    DBus-backed implementation."""

    def screenshot(self, include_pointer, on_done):
        _deliver(on_done, _grab_root())

    def screenshot_area(self, x, y, w, h, include_pointer, on_done):
        _deliver(on_done, _grab_root(x, y, w, h))



#!/usr/bin/python3

import subprocess
import os
import json
import sys

CLI = sys.argv

def parse_args(command):
    return command.split(' ')

def spawn(command):
    try:
        process = subprocess.run(
            parse_args(command),
            stdout=subprocess.PIPE,
            check=True
        )
    except Exception:
        raise subprocess.CalledProcessError(1, command)
    out = process.stdout.decode('utf-8')
    return out

# Utility script that creates GDesktop files for Wine and other window backed applications.
def handle_cli():

    if CLI[1] == 'get_process':
        process = spawn('cat /proc/{}/cmdline'.format(CLI[2]))

        if '.exe' in process:
            if 'Z:' in process:
                process = process.split('Z:')[1]

            process = process.replace('\\', '/')
            process = process.split('.exe')[0] + '.exe'
            process = 'wine '+process.replace(' ', r'\ ')

        process = json.dumps(process)
        if '\\u0000' in process:
            process = process.replace('\\u0000', ' ')
        process = json.loads(process)

        if '.exe' not in process:
            process = process[:-1]

        if process == 'python mainwindow.py':
            process = 'playonlinux'

        try:
            proc_array = process.split('/')
            pa_len = len(proc_array)
            process_name = proc_array[pa_len - 1].title()

            # Since this is a window backed app, make sure it has an icon association.

            icons_dir = '{}/.local/share/icons/hicolor/48x48/apps/'.format(os.getenv('HOME'))

            if '\\ ' in process_name:
                process_name = process_name.replace('\\ ', ' ')

            if '.Exe' in process_name:
                process_name = process_name.replace('.Exe', '')

            icon_file = process_name+'.png'

            if ' ' in icon_file:
                icon_file = icon_file.replace(' ', '')

            icon = icons_dir+icon_file

            try:
                try:
                    spawn('gnome-exe-thumbnailer {} {}'.format(process.split('wine ')[1], icon))
                except IndexError:
                    spawn('gnome-exe-thumbnailer {} {}'.format(process, icon))
            except subprocess.CalledProcessError:
                icon = None

            g_menu = '[Desktop Entry]\n' \
                     'Type=Application\n' \
                     'Encoding=UTF-8\n' \
                     'Name={}\n' \
                     'Comment={}\n' \
                     'Exec={}\n' \
                     'Terminal=false\n' \
                     'StartupNotify=true\n'.format(process_name, process_name, process)

            if icon:
                g_menu += 'Icon={}\n'.format(icon)


            if '.exe' in process:
                g_menu += 'GenericName=Wine application\n' \
                          'Categories=Wine;\n' \
                          'MimeType=application/x-ms-dos-executable;' \
                          'application/x-msi;application/x-ms-shortcut; \n' \

            desktop_file = '{}.cinnamon-generated.desktop'.format(process_name)
            desktop_path = '{}/.local/share/applications/{}'.format(os.getenv('HOME'), desktop_file)

            with open(desktop_path, 'w', encoding='utf-8') as desktop:
                print(g_menu)
                desktop.write(g_menu)
                spawn('chmod +x {}'.format(desktop_path))
                print(desktop_file)

        except KeyError as err:
            print(err)
            return

handle_cli()



import os

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('XApp', '1.0')
from gi.repository import Gdk, Gio, GLib, Gtk, XApp

import prefs
import util


DELAY_VALUES = (0, 2, 5, 10)

PKGDATADIR = '/usr/share/cinnamon/cinnamon-screenshot'
ICON_DIR = os.path.join(PKGDATADIR, 'icons')

Gtk.IconTheme.get_default().append_search_path(ICON_DIR)


class _CropState:
    def __init__(self):
        self.start = None
        self.current = None
        self.dragging = False


class MainWindow:
    def __init__(self, app):
        self.app = app
        self.builder = Gtk.Builder.new_from_file(os.path.join(PKGDATADIR, 'main-window.ui'))
        self.window = self.builder.get_object('window')

        self.mode_screen = self.builder.get_object('mode_screen')
        self.mode_monitor = self.builder.get_object('mode_monitor')
        self.mode_window = self.builder.get_object('mode_window')
        self.mode_area = self.builder.get_object('mode_area')
        self.options_button = self.builder.get_object('options_button')
        self.pointer_switch = self.builder.get_object('pointer_switch')
        self.shadow_switch = self.builder.get_object('shadow_switch')
        self.delay_radios = {v: self.builder.get_object(f'delay_{v}') for v in DELAY_VALUES}
        self.delay_label = self.builder.get_object('delay_label')
        self.take_button = self.builder.get_object('take_button')

        self.preview_stack = self.builder.get_object('preview_stack')
        self.preview_area = self.builder.get_object('preview_area')

        self.crop_button = self.builder.get_object('crop_button')
        self.undo_button = self.builder.get_object('undo_button')
        self.copy_button = self.builder.get_object('copy_button')
        self.save_button = self.builder.get_object('save_button')

        self.service_infobar = self.builder.get_object('service_infobar')

        # Widgets that only work when the cinnamon screenshot service is
        # owned. The X11 fallback can't pick specific windows or arbitrary
        # regions, and can't overlay the cursor or apply window shadows.
        self.service_widgets = XApp.VisibilityGroup.new(True, True, [
            self.mode_window,
            self.mode_area,
            self.pointer_switch,
            self.shadow_switch,
        ])

        self._pixbuf = None
        self._undo_stack = []
        self._suggested_path = None
        self._crop = _CropState()
        self._render_rect = (0, 0, 0, 0, 1.0)
        self._capture_timeout_id = 0
        self._saved = False

        self.window.connect('destroy', self._on_window_destroy)

        self._init_options()
        self._init_preview()
        self._init_actions()
        self._init_service_monitor()
        self._update_action_sensitivity()

    def run(self):
        if self.app.args.interactive:
            self.window.show_all()
        else:
            self._capture(initial=True)

    def show_with_pixbuf(self, pixbuf):
        self._set_preview(pixbuf)
        self.window.show_all()

    # ------------------------------------------------------------------
    # init
    # ------------------------------------------------------------------

    def _init_options(self):
        self.mode_monitor.set_sensitive(Gdk.Display.get_default().get_n_monitors() > 1)

        args = self.app.args
        saved_delay = prefs.get_delay()
        if saved_delay in self.delay_radios:
            self.delay_radios[saved_delay].set_active(True)
        prefs.settings.bind(prefs.INCLUDE_POINTER_KEY, self.pointer_switch, 'active',
                            Gio.SettingsBindFlags.DEFAULT)
        prefs.settings.bind(prefs.INCLUDE_SHADOW_KEY, self.shadow_switch, 'active',
                            Gio.SettingsBindFlags.DEFAULT)
        if args.window or args.select_window:
            self.mode_window.set_active(True)
        elif args.area:
            self.mode_area.set_active(True)
        elif args.monitor is not None and self.mode_monitor.get_sensitive():
            self.mode_monitor.set_active(True)

        for radio in self.delay_radios.values():
            radio.connect('toggled', self._on_delay_toggled)
        self.take_button.connect('clicked', self._on_take_clicked)
        self._on_delay_toggled(None)

    def _init_preview(self):
        self.preview_area.add_events(
            Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
        )
        self.preview_area.connect('realize', self._on_preview_realize)
        self.preview_area.connect('draw', self._on_draw)
        self.preview_area.connect('button-press-event', self._on_press)
        self.preview_area.connect('motion-notify-event', self._on_motion)
        self.preview_area.connect('button-release-event', self._on_release)

        self.crop_button.connect('clicked', self._on_crop)
        self.undo_button.connect('clicked', self._on_undo)
        self.copy_button.connect('clicked', self._on_copy)

        accel = Gtk.AccelGroup()
        accel.connect(Gdk.KEY_z, Gdk.ModifierType.CONTROL_MASK,
                      Gtk.AccelFlags.VISIBLE,
                      lambda *_: self._on_undo(None) or True)
        accel.connect(Gdk.KEY_Escape, 0, Gtk.AccelFlags.VISIBLE,
                      lambda *_: self._on_cancel(None) or True)
        self.window.add_accel_group(accel)

    def _on_preview_realize(self, widget):
        widget.get_window().set_cursor(
            Gdk.Cursor.new_for_display(
                Gdk.Display.get_default(), Gdk.CursorType.CROSSHAIR))

    def _init_service_monitor(self):
        self.app.backend.connect('online-changed', self._on_service_online_changed)
        self._update_service_state(self.app.backend.is_available())

    def _on_service_online_changed(self, _backend, online):
        self._update_service_state(online)

    def _update_service_state(self, online):
        self.service_infobar.set_revealed(not online)
        self.mode_monitor.set_sensitive(Gdk.Display.get_default().get_n_monitors() > 1)
        self.service_widgets.set_sensitive(online)
        if not online and (self.mode_window.get_active() or self.mode_area.get_active()):
            self.mode_screen.set_active(True)

    def _init_actions(self):
        self.save_button.connect('clicked', self._on_save)
        self.builder.get_object('cancel_button').connect('clicked', self._on_cancel)

        menu = Gtk.Menu()
        open_folder_item = Gtk.MenuItem(label=_('Open save folder'))
        open_folder_item.connect('activate', self._on_open_save_folder)
        menu.append(open_folder_item)
        menu.append(Gtk.SeparatorMenuItem())
        prefs_item = Gtk.MenuItem(label=_('Preferences'))
        prefs_item.connect('activate', self._on_preferences)
        menu.append(prefs_item)
        menu.show_all()
        self.builder.get_object('menu_button').set_popup(menu)

    # ------------------------------------------------------------------
    # mode handling and capture
    # ------------------------------------------------------------------

    def _on_delay_toggled(self, radio):
        delay = self._selected_delay()
        self.delay_label.set_label(f'{delay}s')
        if radio is not None and radio.get_active():
            prefs.set_delay(delay)

    def _selected_delay(self):
        for v, radio in self.delay_radios.items():
            if radio.get_active():
                return v
        return 0

    def _on_take_clicked(self, _b):
        self._capture()

    def _capture(self, initial=False):
        if self.mode_window.get_active():
            mode = 'window' if initial and self.app.args.window else 'select_window'
        elif self.mode_area.get_active():
            mode = 'area'
        elif self.mode_monitor.get_active():
            mode = 'monitor'
        else:
            mode = 'screen'

        if initial:
            args = self.app.args
            delay = args.delay if args.delay is not None else 0
            include_pointer = args.include_pointer
            include_shadow = args.include_shadow
        else:
            delay = self._selected_delay()
            include_pointer = self.pointer_switch.get_active()
            include_shadow = self.shadow_switch.get_active()

        area_rect = None
        if mode == 'monitor':
            if initial and self.app.args.monitor is not None:
                area_rect = util.monitor_rect(self.app.args.monitor)
            else:
                area_rect = util.monitor_rect_for_window(self.window.get_window())
            if area_rect is None:
                mode = 'screen'

        def done(pixbuf):
            self._set_preview(pixbuf)
            self.window.show_all()

        def start_capture():
            self._capture_timeout_id = 0
            self.app.capture(mode, include_pointer, include_shadow, delay, done, area_rect=area_rect)
            return GLib.SOURCE_REMOVE

        if self.window.get_visible():
            self.window.hide()
            self._capture_timeout_id = GLib.timeout_add(200, start_capture)
        else:
            start_capture()

    def _set_preview(self, pixbuf):
        if pixbuf is None:
            self._update_action_sensitivity()
            return
        self._pixbuf = pixbuf
        self._undo_stack.clear()
        self._crop.start = None
        self._crop.current = None
        self._crop.dragging = False
        self._suggested_path = util.build_filename(
            prefs.get_save_directory(),
            file_type=prefs.get_default_file_type(),
        )
        self._update_action_sensitivity()
        self.preview_area.queue_draw()

    def _update_action_sensitivity(self):
        has_preview = self._pixbuf is not None
        has_selection = self._current_selection_widget_coords() is not None
        self.preview_stack.set_visible_child_name('preview' if has_preview else 'placeholder')
        self.crop_button.set_sensitive(has_preview and has_selection)
        self.copy_button.set_sensitive(has_preview)
        self.save_button.set_sensitive(has_preview)
        self.undo_button.set_sensitive(has_preview and bool(self._undo_stack))

    # ------------------------------------------------------------------
    # preview rendering
    # ------------------------------------------------------------------

    def _on_draw(self, widget, cr):
        if self._pixbuf is None:
            return False
        alloc = widget.get_allocation()
        pw, ph = self._pixbuf.get_width(), self._pixbuf.get_height()
        if pw == 0 or ph == 0:
            return False

        scale = min(alloc.width / pw, alloc.height / ph, 1.0)
        rw, rh = pw * scale, ph * scale
        rx = (alloc.width - rw) / 2
        ry = (alloc.height - rh) / 2
        self._render_rect = (rx, ry, rw, rh, scale)

        cr.save()
        cr.translate(rx, ry)
        cr.scale(scale, scale)
        Gdk.cairo_set_source_pixbuf(cr, self._pixbuf, 0, 0)
        cr.paint()
        cr.restore()

        self._draw_crop_overlay(cr, alloc)

        return False

    def _draw_crop_overlay(self, cr, alloc):
        sel = self._current_selection_widget_coords()
        if sel is None:
            return
        sx, sy, sw, sh = sel
        cr.save()
        cr.set_line_width(1.0)
        cr.rectangle(sx + 0.5, sy + 0.5, sw, sh)
        cr.set_source_rgb(1.0, 1.0, 1.0)
        cr.stroke_preserve()
        cr.set_source_rgb(0.0, 0.0, 0.0)
        cr.set_dash([4.0, 4.0])
        cr.stroke()
        cr.restore()

    def _current_selection_widget_coords(self):
        if self._crop.start is None or self._crop.current is None:
            return None
        x1, y1 = self._crop.start
        x2, y2 = self._crop.current
        w = abs(x2 - x1)
        h = abs(y2 - y1)
        if w < 1 or h < 1:
            return None
        return (min(x1, x2), min(y1, y2), w, h)

    # ------------------------------------------------------------------
    # crop interaction
    # ------------------------------------------------------------------

    def _on_press(self, _w, event):
        if self._pixbuf is None or event.button != 1:
            return False
        self._crop.start = (event.x, event.y)
        self._crop.current = (event.x, event.y)
        self._crop.dragging = True
        self.preview_area.queue_draw()
        self._update_action_sensitivity()
        return True

    def _on_motion(self, _w, event):
        if not self._crop.dragging:
            return False
        self._crop.current = (event.x, event.y)
        self.preview_area.queue_draw()
        return True

    def _on_release(self, _w, event):
        if not self._crop.dragging or event.button != 1:
            return False
        self._crop.current = (event.x, event.y)
        self._crop.dragging = False
        self.preview_area.queue_draw()
        self._update_action_sensitivity()
        return True

    def _on_crop(self, _button):
        sel = self._current_selection_widget_coords()
        if self._pixbuf is None or sel is None:
            return
        sx, sy, sw, sh = sel
        rx, ry, _rw, _rh, scale = self._render_rect
        if scale <= 0:
            return
        self._crop.start = None
        self._crop.current = None

        pw = self._pixbuf.get_width()
        ph = self._pixbuf.get_height()
        x = max(0, min(int((sx - rx) / scale), pw))
        y = max(0, min(int((sy - ry) / scale), ph))
        w = max(1, min(int(sw / scale), pw - x))
        h = max(1, min(int(sh / scale), ph - y))

        self._undo_stack.append(self._pixbuf)
        self._pixbuf = self._pixbuf.new_subpixbuf(x, y, w, h).copy()
        self._update_action_sensitivity()
        self.preview_area.queue_draw()

    def _on_undo(self, _button):
        if not self._undo_stack:
            return
        self._pixbuf = self._undo_stack.pop()
        self._update_action_sensitivity()
        self.preview_area.queue_draw()

    # ------------------------------------------------------------------
    # action buttons
    # ------------------------------------------------------------------

    def _on_cancel(self, _b):
        self.window.destroy()
        self.app.quit()

    def _on_window_destroy(self, _w):
        if self._capture_timeout_id != 0:
            GLib.source_remove(self._capture_timeout_id)
            self._capture_timeout_id = 0
        # When the caller (e.g. xdg-desktop-portal) pinned a destination
        # via --file, exit non-zero if the user closed the window without
        # saving so the caller can distinguish a save from a cancel.
        if self.app.args.file and not self._saved:
            self.app._exit_code = 1
        self.app.quit()

    def _on_copy(self, _b):
        if self._pixbuf is None:
            return
        util.copy_pixbuf_to_clipboard(self._pixbuf)

    def _on_save(self, _b):
        if self._pixbuf is None:
            return

        # When --file is given alongside --interactive, the caller has
        # pre-chosen the destination (e.g. xdg-desktop-portal); skip the
        # chooser and save straight to that path.
        if self.app.args.file:
            path = self.app.args.file
        else:
            dialog = Gtk.FileChooserNative.new(
                _('Save Screenshot'),
                self.window,
                Gtk.FileChooserAction.SAVE,
                _('Save'),
                _('Cancel'),
            )
            dialog.set_do_overwrite_confirmation(True)
            dialog.set_current_name(os.path.basename(self._suggested_path))
            suggested_dir = os.path.dirname(self._suggested_path)
            if suggested_dir and os.path.isdir(suggested_dir):
                dialog.set_current_folder(suggested_dir)

            response = dialog.run()
            path = dialog.get_filename() if response == Gtk.ResponseType.ACCEPT else None
            dialog.destroy()
            if not path:
                return

        try:
            util.save_pixbuf(self._pixbuf, path)
        except Exception as exc:
            err = Gtk.MessageDialog(
                transient_for=self.window,
                modal=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text=_('Failed to save screenshot'),
                secondary_text=str(exc),
            )
            err.run()
            err.destroy()
            return

        self._saved = True

        if prefs.get_autosave_to_clipboard():
            util.copy_pixbuf_to_clipboard(self._pixbuf)

        # When --file is set the destination wasn't user-chosen (portal use),
        # so the post-save file-manager reveal would be confusing.
        if prefs.get_launch_file_manager() and not self.app.args.file:
            util.show_in_file_manager(Gio.File.new_for_path(path).get_uri())

        self.window.destroy()
        self.app.quit()

    def _on_open_save_folder(self, _item):
        uri = prefs.get_save_directory_uri()
        if not uri:
            uri = Gio.File.new_for_path(prefs.get_save_directory()).get_uri()
        util.show_in_file_manager(uri)

    def _on_preferences(self, _item):
        prefs.open_preferences(self.window)



#!/usr/bin/python3

try:
    import pam
    print("Using pam module (python3-pampy)")
except:
    import PAM
    print("Using PAM module (python3-pam)")
    pam = None
import pexpect
import time
from random import randint
import shutil
import os
import subprocess

from PIL import Image
import gi
gi.require_version('AccountsService', '1.0')
from gi.repository import AccountsService, GLib, GdkPixbuf, XApp

from bin.SettingsWidgets import SidePage
from bin.ChooserButtonWidgets import PictureChooserButton
from xapp.GSettingsWidgets import *

class PasswordError(Exception):
    """Exception raised when an incorrect password is supplied."""
    pass


class Module:
    name = "user"
    category = "prefs"
    comment = _("Change your user preferences and password")

    def __init__(self, content_box):
        keywords = _("user, account, information, details, password")
        sidePage = SidePage(_("Account Details"), "cs-user", keywords, content_box, module=self)
        self.sidePage = sidePage
        self.window = None

    def _setParentRef(self, window):
        self.window = window

    def on_module_selected(self):
        if not self.loaded:
            print("Loading User module")

            page = SettingsPage()
            self.sidePage.add_widget(page)

            settings = page.add_section(_("Account Details"))

            self.scale = self.window.get_scale_factor()

            self.face_button = PictureChooserButton(num_cols=4, button_picture_width=64, menu_picture_width=64*self.scale, keep_square=True)
            self.face_button.set_alignment(0.0, 0.5)
            self.face_button.set_tooltip_text(_("Click to change your picture"))

            self.face_photo_menuitem = Gtk.MenuItem.new_with_label(_("Take a photo..."))
            self.face_photo_menuitem.connect('activate', self._on_face_photo_menuitem_activated)

            self.face_browse_menuitem = Gtk.MenuItem.new_with_label(_("Browse for more pictures..."))
            self.face_browse_menuitem.connect('activate', self._on_face_browse_menuitem_activated)

            face_dirs = ["/usr/share/cinnamon/faces"]
            for face_dir in face_dirs:
                if os.path.exists(face_dir):
                    pictures = sorted(os.listdir(face_dir))
                    for picture in pictures:
                        path = os.path.join(face_dir, picture)
                        self.face_button.add_picture(path, self._on_face_menuitem_activated)

            widget = SettingsWidget()
            label = Gtk.Label.new(_("Picture"))
            widget.pack_start(label, False, False, 0)
            widget.pack_end(self.face_button, False, False, 0)
            settings.add_row(widget)

            size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.HORIZONTAL)

            widget = SettingsWidget()
            label = Gtk.Label.new(_("Name"))
            widget.pack_start(label, False, False, 0)
            self.realname_entry = EditableEntry()
            size_group.add_widget(self.realname_entry)
            self.realname_entry.connect("changed", self._on_realname_changed)
            self.realname_entry.set_tooltip_text(_("Click to change your name"))
            widget.pack_end(self.realname_entry, False, False, 0)
            settings.add_row(widget)

            widget = SettingsWidget()
            label = Gtk.Label.new(_("Password"))
            widget.pack_start(label, False, False, 0)
            password_mask = Gtk.Label.new('\u2022\u2022\u2022\u2022\u2022\u2022')
            password_mask.set_alignment(0.9, 0.5)
            self.password_button = Gtk.Button()
            size_group.add_widget(self.password_button)
            self.password_button.add(password_mask)
            self.password_button.set_relief(Gtk.ReliefStyle.NONE)
            self.password_button.set_tooltip_text(_("Click to change your password"))
            self.password_button.connect('activate', self._on_password_button_clicked)
            self.password_button.connect('released', self._on_password_button_clicked)
            widget.pack_end(self.password_button, False, False, 0)
            settings.add_row(widget)

            current_user = GLib.get_user_name()
            self.accountService = AccountsService.UserManager.get_default().get_user(current_user)
            self.accountService.connect('notify::is-loaded', self.load_user_info)

            self.face_button.add_separator()

            # Video devices assumed to be webcams
            import glob
            webcam_detected = len(glob.glob("/dev/video*")) > 0

            if webcam_detected:
                self.face_button.add_menuitem(self.face_photo_menuitem)

            self.face_button.add_menuitem(self.face_browse_menuitem)

    def update_preview_cb (self, dialog, preview):
        filename = dialog.get_preview_filename()
        if filename is not None:
            if os.path.isfile(filename):
                try:
                    pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(filename, 128, 128)
                    if pixbuf is not None:
                        preview.set_from_pixbuf(pixbuf)
                        self.frame.show()
                        return
                except GLib.Error as e:
                    print(f"Unable to generate preview for file '{filename}' - {e.message}\n")

        preview.clear()
        self.frame.hide()

    def _on_face_photo_menuitem_activated(self, menuitem):

        # streamer takes -t photos, uses /dev/video0
        if 0 != subprocess.call(["streamer", "-j90", "-t8", "-s800x600", "-o", "/tmp/temp-account-pic00.jpeg"]):
            print("Error: Webcam not available")
            return

        # Use the 8th frame (the webcam takes a few frames to "lighten up")
        path = "/tmp/temp-account-pic07.jpeg"

        # Crop the image to thumbnail size
        image = Image.open(path)
        width, height = image.size

        if width > height:
            new_width = height
            new_height = height
        elif height > width:
            new_width = width
            new_height = width
        else:
            new_width = width
            new_height = height

        left = (width - new_width) / 2
        top = (height - new_height) / 2
        right = (width + new_width) / 2
        bottom = (height + new_height) / 2

        image = image.crop((left, top, right, bottom))
        image.thumbnail((255, 255), Image.LANCZOS)

        face_path = os.path.join(self.accountService.get_home_dir(), ".face")

        image.save(face_path, "png")
        self.accountService.set_icon_file(face_path)
        self.face_button.set_picture_from_file(face_path)


    def _on_face_browse_menuitem_activated(self, menuitem):
        dialog = Gtk.FileChooserDialog(None, None, Gtk.FileChooserAction.OPEN, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        dialog.set_current_folder(self.accountService.get_home_dir())
        filter = Gtk.FileFilter()
        filter.set_name(_("Images"))
        filter.add_mime_type("image/*")
        dialog.add_filter(filter)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.frame = Gtk.Frame(visible=False, no_show_all=True)
        preview = Gtk.Image(visible=True)

        box.pack_start(self.frame, False, False, 0)
        self.frame.add(preview)
        dialog.set_preview_widget(box)
        dialog.set_preview_widget_active(True)
        dialog.set_use_preview_label(False)

        box.set_margin_end(12)
        box.set_margin_top(12)
        box.set_size_request(128, -1)

        dialog.connect("update-preview", self.update_preview_cb, preview)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            string = dialog.get_filename()
            print(string)
            if string.startswith("/"):
                path = string
            else:
                theme = Gtk.IconTheme.get_default()
                icon_info = theme.lookup_icon_for_scale(string, 256, dialog.get_scale_factor(), Gtk.IconLookupFlags.FORCE_SIZE)
                path = icon_info.get_filename() if icon_info else None

            face_path = os.path.join(self.accountService.get_home_dir(), ".face")

            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, 255, -1)
            pixbuf.savev(face_path, "png")
            self.accountService.set_icon_file(path)
            self.face_button.set_picture_from_file(path)

        dialog.destroy()

    def _on_face_menuitem_activated(self, path):
        if os.path.exists(path):
            self.accountService.set_icon_file(path)
            shutil.copy(path, os.path.join(self.accountService.get_home_dir(), ".face"))
            return True

    def load_user_info(self, user, param):
        self.realname_entry.set_text(user.get_real_name())
        for path in [os.path.join(self.accountService.get_home_dir(), ".face"), user.get_icon_file(), "/usr/share/cinnamon/faces/user-generic.png"]:
            if os.path.exists(path):
                self.face_button.set_picture_from_file(path)
                break

    def _on_realname_changed(self, widget, text):
        self.accountService.set_real_name(text)

    def _on_password_button_clicked(self, widget):
        dialog = PasswordDialog()
        response = dialog.run()


class PasswordDialog(Gtk.Dialog):

    def __init__ (self):
        super(PasswordDialog, self).__init__()

        self.correct_current_password = False # Flag to remember if the current password is correct or not

        self.set_modal(True)
        self.set_skip_taskbar_hint(True)
        self.set_skip_pager_hint(True)
        self.set_title(_("Change Password"))

        table = Gtk.Table(6, 3)
        table.set_border_width(6)
        table.set_row_spacings(8)
        table.set_col_spacings(15)

        label = Gtk.Label.new(_("Current password"))
        label.set_alignment(1, 0.5)
        table.attach(label, 0, 1, 0, 1)

        label = Gtk.Label.new(_("New password"))
        label.set_alignment(1, 0.5)
        table.attach(label, 0, 1, 1, 2)

        label = Gtk.Label.new(_("Confirm password"))
        label.set_alignment(1, 0.5)
        table.attach(label, 0, 1, 3, 4)

        self.current_password = Gtk.Entry()
        self.current_password.set_visibility(False)
        self.current_password.connect("focus-out-event", self._on_current_password_changed)
        table.attach(self.current_password, 1, 3, 0, 1)

        self.new_password = Gtk.Entry()
        self.new_password.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "view-refresh")
        self.new_password.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("Generate a password"))
        self.new_password.set_tooltip_text(_("Generate a password"))
        self.new_password.connect("icon-release", self._on_new_password_icon_released)
        self.new_password.connect("changed", self._on_passwords_changed)
        table.attach(self.new_password, 1, 3, 1, 2)

        self.strengh_indicator = Gtk.ProgressBar()
        self.strengh_indicator.set_tooltip_text(_("Your new password needs to be at least 8 characters long"))
        self.strengh_indicator.set_fraction(0.0)
        table.attach(self.strengh_indicator, 1, 2, 2, 3, xoptions=Gtk.AttachOptions.EXPAND|Gtk.AttachOptions.FILL)
        self.strengh_indicator.set_size_request(-1, 1)

        self.strengh_label = Gtk.Label()
        self.strengh_label.set_tooltip_text(_("Your new password needs to be at least 8 characters long"))
        self.strengh_label.set_alignment(1, 0.5)
        table.attach(self.strengh_label, 2, 3, 2, 3)

        self.confirm_password = Gtk.Entry()
        self.confirm_password.connect("changed", self._on_passwords_changed)
        table.attach(self.confirm_password, 1, 3, 3, 4)

        self.show_password = Gtk.CheckButton(_("Show password"))
        self.show_password.connect('toggled', self._on_show_password_toggled)
        table.attach(self.show_password, 1, 3, 4, 5)

        self.set_border_width(6)

        box = self.get_content_area()
        box.add(table)
        self.show_all()

        self.infobar = Gtk.InfoBar()
        self.infobar.set_message_type(Gtk.MessageType.ERROR)
        label = Gtk.Label.new(_("An error occurred. Your password was not changed."))
        content = self.infobar.get_content_area()
        content.add(label)
        table.attach(self.infobar, 0, 3, 5, 6)

        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, _("Change"), Gtk.ResponseType.OK, )

        self.set_passwords_visibility()
        self.set_response_sensitive(Gtk.ResponseType.OK, False)
        self.infobar.hide()

        self.connect("response", self._on_response)

    def _on_response(self, dialog, response_id):
        if response_id == Gtk.ResponseType.OK:
            self.change_password()
        else:
            self.destroy()

    def change_password(self):
        oldpass = self.current_password.get_text()
        newpass = self.new_password.get_text()
        passwd = pexpect.spawn("/usr/bin/passwd")
        # passwd only asks for the old password when there already is one set.
        if oldpass == "":
            time.sleep(0.5)
            passwd.sendline(newpass)
            time.sleep(0.5)
            passwd.sendline(newpass)
        else:
            time.sleep(0.5)
            passwd.sendline(oldpass)
            time.sleep(0.5)
            passwd.sendline(newpass)
            time.sleep(0.5)
            passwd.sendline(newpass)
        time.sleep(0.5)
        passwd.close()

        if passwd.exitstatus is None or passwd.exitstatus > 0:
            self.infobar.show_all()
        else:
            self.destroy()

    def set_passwords_visibility(self):
        visible = self.show_password.get_active()
        self.new_password.set_visibility(visible)
        self.confirm_password.set_visibility(visible)

    def _on_new_password_icon_released(self, widget, icon_pos, event):
        self.infobar.hide()
        self.show_password.set_active(True)
        characters = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-"
        newpass = ""
        for i in range (8):
            index = randint(0, len(characters) -1)
            newpass = newpass + characters[index]

        self.new_password.set_text(newpass)
        self.confirm_password.set_text(newpass)
        self.check_passwords()

    def _on_show_password_toggled(self, widget):
        self.set_passwords_visibility()

    def _get_pam_service(self):
        import os
        if os.path.exists('/etc/pam.d/system-auth'):
            return 'system-auth'
        elif os.path.exists('/etc/pam.d/common-auth'):
            return 'common-auth'
        else:
            return 'login'

    def auth_pam(self):
        service = self._get_pam_service()
        if not pam.pam().authenticate(GLib.get_user_name(), self.current_password.get_text(), service):
            raise PasswordError("Invalid password")

    def auth_PyPAM(self):
        auth = PAM.pam()
        auth.start(self._get_pam_service())
        auth.set_item(PAM.PAM_USER, GLib.get_user_name())
        auth.set_item(PAM.PAM_CONV, self.pam_conv)
        try:
            auth.authenticate()
            auth.acct_mgmt()
            return True
        except PAM.error as resp:
            raise PasswordError("Invalid password")

    def _on_current_password_changed(self, widget, event):
        self.infobar.hide()
        try:
            self.auth_pam() if pam else self.auth_PyPAM()
        except PasswordError:
            self.current_password.set_icon_from_stock(Gtk.EntryIconPosition.SECONDARY, Gtk.STOCK_DIALOG_WARNING)
            self.current_password.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("Wrong password"))
            self.current_password.set_tooltip_text(_("Wrong password"))
            self.correct_current_password = False
        except:
            self.current_password.set_icon_from_stock(Gtk.EntryIconPosition.SECONDARY, Gtk.STOCK_DIALOG_WARNING)
            self.current_password.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("Internal Error"))
            self.current_password.set_tooltip_text(_("Internal Error"))
            self.correct_current_password = False
            raise
        else:
            self.current_password.set_icon_from_stock(Gtk.EntryIconPosition.SECONDARY, None)
            self.current_password.set_tooltip_text("")
            self.correct_current_password = True
            self.check_passwords()

    # Based on setPasswordStrength() in Mozilla Seamonkey, which is tri-licensed under MPL 1.1, GPL 2.0, and LGPL 2.1.
    # Forked from Ubiquity validation.py
    def password_strength(self, password):
        upper = lower = digit = symbol = 0
        for char in password:
            if char.isdigit():
                digit += 1
            elif char.islower():
                lower += 1
            elif char.isupper():
                upper += 1
            else:
                symbol += 1
        length = len(password)

        length = min(length,4)
        digit = min(digit,3)
        upper = min(upper,3)
        symbol = min(symbol,3)
        strength = (
            ((length * 0.1) - 0.2) +
            (digit * 0.1) +
            (symbol * 0.15) +
            (upper * 0.1))
        if strength > 1:
            strength = 1
        if strength < 0:
            strength = 0
        return strength

    def _on_passwords_changed(self, widget):
        self.infobar.hide()
        new_password = self.new_password.get_text()
        confirm_password = self.confirm_password.get_text()
        strength = self.password_strength(new_password)
        if new_password != confirm_password:
            self.confirm_password.set_icon_from_stock(Gtk.EntryIconPosition.SECONDARY, Gtk.STOCK_DIALOG_WARNING)
            self.confirm_password.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("Passwords do not match"))
            self.confirm_password.set_tooltip_text(_("Passwords do not match"))
        else:
            self.confirm_password.set_icon_from_stock(Gtk.EntryIconPosition.SECONDARY, None)
            self.confirm_password.set_tooltip_text("")
        if len(new_password) < 8:
            self.strengh_label.set_text(_("Too short"))
            self.strengh_indicator.set_fraction(0.0)
        elif strength < 0.6:
            self.strengh_label.set_text(_("Weak"))
            self.strengh_indicator.set_fraction(0.2)
        elif strength < 0.75:
            self.strengh_label.set_text(_("Fair"))
            self.strengh_indicator.set_fraction(0.4)
        elif strength < 0.9:
            self.strengh_label.set_text(_("Good"))
            self.strengh_indicator.set_fraction(0.6)
        else:
            self.strengh_label.set_text(_("Strong"))
            self.strengh_indicator.set_fraction(1.0)

        self.check_passwords()

    def check_passwords(self):
        if self.correct_current_password:
            new_password = self.new_password.get_text()
            confirm_password = self.confirm_password.get_text()
            if len(new_password) >= 8 and new_password == confirm_password:
                self.set_response_sensitive(Gtk.ResponseType.OK, True)
            else:
                self.set_response_sensitive(Gtk.ResponseType.OK, False)

    def pam_conv(self, auth, query_list, userData):
        resp = []
        for i in range(len(query_list)):
            query, type = query_list[i]
            val = self.current_password.get_text()
            resp.append((val, 0))
        return resp



class Backend:
    """Backends deliver results via on_done(result) rather than returning
    them, so DBus-backed implementations can run without blocking the UI."""

    def screenshot(self, include_pointer, on_done):
        on_done(None)

    def screenshot_window(self, include_pointer, include_shadow, on_done):
        on_done(None)

    def screenshot_window_by_id(self, window_id, include_pointer, include_shadow, on_done):
        on_done(None)

    def screenshot_area(self, x, y, w, h, include_pointer, on_done):
        on_done(None)

    def flash_area(self, x, y, w, h):
        pass

    def select_area(self, on_done):
        on_done(None)

    def select_window(self, on_done):
        on_done(None)



#!/usr/bin/python3

import os
import subprocess

import cairo
import gi
gi.require_version("Gtk", "3.0")
gi.require_version("CinnamonDesktop", "3.0")
gi.require_version('IBus', '1.0')
from gi.repository import GLib, Gio, Gtk, GObject, CinnamonDesktop, IBus

from bin.SettingsWidgets import GSettingsKeybinding
from xapp.SettingsWidgets import SettingsPage
from xapp.GSettingsWidgets import PXGSettingsBackend, GSettingsSwitch

from bin import AddKeyboardLayout
from bin.util import using_fcitx

MAX_LAYOUTS_PER_GROUP = 4

class InputSourceSettingsPage(SettingsPage):
    def __init__(self):
        super().__init__()

        self._reselect_index = None

        builder = Gtk.Builder()
        builder.set_translation_domain('cinnamon')
        builder.add_from_file("/usr/share/cinnamon/cinnamon-settings/bin/input-sources-list.ui")

        box = builder.get_object("input_sources_list_box")
        self.pack_start(box, False, False, 0)

        self.input_sources_list = builder.get_object("input_sources_list")
        self.source_activate_handler = self.input_sources_list.connect("row-activated", self.on_input_source_activated)
        self.current_input_sources_model = CurrentInputSourcesModel()
        self.current_input_sources_model.connect("items-changed", self.on_model_updated)
        self.input_sources_list.bind_model(self.current_input_sources_model, self.current_input_sources_model.create_row)
        self.input_sources_list.set_header_func(self.row_separator_func)

        self.test_layout_button = builder.get_object("test_layout")
        self.test_layout_button.connect("clicked", self.on_test_layout_clicked)
        # TODO: maybe use tecla as an alternative for wayland, if we don't roll something ourselves.
        # btw there's no plan for tecla to support different keyboard geometries than a standard pc105.
        if not GLib.find_program_in_path("gkbd-keyboard-display"):
            self.test_layout_button.set_visible(False)

        self.engine_config_button = builder.get_object("engine_config_button")
        self.engine_config_button.connect("clicked", self.on_engine_config_clicked)

        self.im_launch_button = builder.get_object("im_launch_button")
        self.im_launch_button.connect("clicked", lambda widget: subprocess.Popen(["mintlocale-im"]))
        if not GLib.find_program_in_path("mintlocale-im"):
            self.im_launch_button.set_visible(False)

        # Configure is ibus-engine-specific; fcitx engines are configured in fcitx.
        if using_fcitx():
            self.engine_config_button.set_no_show_all(True)
            self.engine_config_button.set_visible(False)
        self.add_layout_button = builder.get_object("add_layout")
        self.add_layout_button.connect("clicked", self.on_add_layout_clicked)
        self.remove_layout_button = builder.get_object("remove_layout")
        self.remove_layout_button.connect("clicked", self.on_remove_layout_clicked)
        self.move_layout_up_button = builder.get_object("move_layout_up")
        self.move_layout_up_button.connect("clicked", self.on_move_layout_up_clicked)
        self.move_layout_down_button = builder.get_object("move_layout_down")
        self.move_layout_down_button.connect("clicked", self.on_move_layout_down_clicked)

        self.update_widgets()

        if using_fcitx():
            # Under fcitx, layout switching, the applet and per-window memory are
            # owned by fcitx, and input methods are configured in fcitx's own
            # tools. Explain that, with handoffs to those tools, in place of our
            # options/shortcuts.
            infobar = Gtk.InfoBar()
            infobar.set_message_type(Gtk.MessageType.INFO)
            label = Gtk.Label(
                _("In Fcitx mode, input methods, additional layouts and their "
                  "switching shortcuts are managed by Fcitx. Only the first "
                  "layout configured here is used.")
            )
            label.set_line_wrap(True)
            label.set_xalign(0.0)
            infobar.get_content_area().add(label)

            manage_button = Gtk.Button(label=_("Manage Fcitx"))
            manage_button.connect("clicked", lambda widget: subprocess.Popen(["fcitx5-configtool"]))
            infobar.get_action_area().pack_start(manage_button, False, False, 0)

            infobar.show_all()
            self.pack_start(infobar, False, False, 0)
            self.reorder_child(infobar, 0)
            return

        section = self.add_section(_("Options"))
        widget = GSettingsSwitch(
            _("Remember the last layout used for each window"),
            "org.cinnamon.desktop.input-sources", "per-window"
        )
        section.add_row(widget)

        widget = GSettingsSwitch(
            _("Use a country flag, if available, to represent keyboard layouts"),
            "org.cinnamon.desktop.interface", "keyboard-layout-show-flags"
        )
        section.add_row(widget)
        widget = GSettingsSwitch(
            _("Use the layout name, not the group name, to represent a layout"),
            "org.cinnamon.desktop.interface", "keyboard-layout-prefer-variant-names"
        )
        section.add_row(widget)
        widget = GSettingsSwitch(
            _("Use upper-case when using text to represent a layout"),
            "org.cinnamon.desktop.interface", "keyboard-layout-use-upper"
        )
        section.add_row(widget)

        section = self.add_section(_("Shortcuts"))
        widget = GSettingsKeybinding(
            _("Switch to next layout"),
            "org.cinnamon.desktop.keybindings.wm",
            "switch-input-source",
        )
        section.add_row(widget)
        widget = GSettingsKeybinding(
            _("Switch to previous layout"),
            "org.cinnamon.desktop.keybindings.wm",
            "switch-input-source-backward"
        )
        section.add_row(widget)
        widget = GSettingsKeybinding(
            _("Switch to first layout"),
            "org.cinnamon.desktop.keybindings.wm",
            "switch-input-source-0"
        )
        section.add_row(widget)
        widget = GSettingsKeybinding(
            _("Switch to second layout"),
            "org.cinnamon.desktop.keybindings.wm",
            "switch-input-source-1"
        )
        section.add_row(widget)
        widget = GSettingsKeybinding(
            _("Switch to third layout"),
            "org.cinnamon.desktop.keybindings.wm",
            "switch-input-source-2"
        )
        section.add_row(widget)
        widget = GSettingsKeybinding(
            _("Switch to fourth layout"),
            "org.cinnamon.desktop.keybindings.wm",
            "switch-input-source-3"
        )
        section.add_row(widget)


    def row_separator_func(self, row, before, data=None):
        if before is None:
            row.set_header(None)
            return

        row.set_header(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))

    def on_model_updated(self, model, position, removed, added, data=None):
        GLib.idle_add(self._update_selected_row)
        self.update_widgets()

    def _update_selected_row(self, data=None):
        self.input_sources_list.handler_block(self.source_activate_handler)

        row = None
        if self._reselect_index is not None:
            row = self.input_sources_list.get_row_at_index(self._reselect_index)
            self._reselect_index = None

        if row is not None:
            self.input_sources_list.select_row(row)
        else:
            self.input_sources_list.unselect_all()

        self.input_sources_list.handler_unblock(self.source_activate_handler)
        self.update_widgets()

        return GLib.SOURCE_REMOVE

    def on_input_source_activated(self, listbox, row):
        self.update_widgets()

    def _get_selected_source(self):
        source = None
        rows = self.input_sources_list.get_selected_rows()
        if len(rows) > 0:
            source = rows[0].get_child().input_source
        return source

    def on_add_layout_clicked(self, button, data=None):
        self.current_input_sources_model.show_add_layout_dialog()

    def on_remove_layout_clicked(self, button, data=None):
        source = self._get_selected_source()
        if source is not None:
            self.current_input_sources_model.remove_layout(source)

    def on_move_layout_up_clicked(self, button, data=None):
        source = self._get_selected_source()
        if source is not None:
            self._reselect_index = self.current_input_sources_model.get_item_index(source) - 1
            self.current_input_sources_model.move_layout_up(source)

    def on_move_layout_down_clicked(self, button, data=None):
        source = self._get_selected_source()
        if source is not None:
            self._reselect_index = self.current_input_sources_model.get_item_index(source) + 1
            self.current_input_sources_model.move_layout_down(source)

    def on_test_layout_clicked(self, button, data=None):
        source = self._get_selected_source()

        args = AddKeyboardLayout.make_gkbd_keyboard_args(source.xkb_layout, source.xkb_variant)
        subprocess.Popen(args)

    def on_engine_config_clicked(self, button, data=None):
        source = self._get_selected_source()
        dialog = IBusConfigDialog(source, self.current_input_sources_model.input_source_settings)
        dialog.run()
        dialog.destroy()

    def update_widgets(self):
        # Don't allow removal of last remaining layout
        n_items = self.current_input_sources_model.get_n_items()
        self.add_layout_button.set_sensitive(n_items < MAX_LAYOUTS_PER_GROUP)

        source = self._get_selected_source()
        if source is not None:
            self.test_layout_button.set_sensitive(source.type == "xkb")
            # Enable Configure button for all IBus sources (not just those with preferences)
            self.engine_config_button.set_sensitive(source.type == "ibus")
            index = self.current_input_sources_model.get_item_index(source)
            self.move_layout_up_button.set_sensitive(index > 0)
            self.move_layout_down_button.set_sensitive(index < self.current_input_sources_model.get_n_items() - 1)
            self.remove_layout_button.set_sensitive(n_items > 1)
        else:
            self.test_layout_button.set_sensitive(False)
            self.engine_config_button.set_sensitive(False)
            self.move_layout_up_button.set_sensitive(False)
            self.move_layout_down_button.set_sensitive(False)
            self.remove_layout_button.set_sensitive(False)


class IBusConfigDialog():
    def __init__(self, source, settings):
        self.source = source
        self.settings = settings
        self.xkb_info = CinnamonDesktop.XkbInfo.new_with_extras()

        # Check if this engine uses "default" layout or has a specific one
        self.engine_layout = self._get_engine_layout()
        self.allows_override = (self.engine_layout == "default")

        builder = Gtk.Builder()
        builder.set_translation_domain('cinnamon')
        builder.add_from_file("/usr/share/cinnamon/cinnamon-settings/bin/input-sources-list.ui")

        self.dialog = builder.get_object("ibus_config_dialog")

        # Set up name label
        name_label = builder.get_object("ibus_config_name_label")
        name_label.set_text(source.display_name)

        # Set up explanation label based on whether override is allowed
        explanation_label = builder.get_object("ibus_config_explanation_label")
        if not self.allows_override:
            layout_display = self._get_engine_layout_display_name()
            explanation_label.set_text(
                _("This input method requires the \"%s\" keyboard layout to function correctly. "
                  "The layout is set automatically when you switch to this input method.") % layout_display
            )

        # Set up layout label
        self.layout_label = builder.get_object("ibus_config_layout_label")

        # Set up buttons
        close_button = builder.get_object("ibus_config_close_button")
        close_button.connect("clicked", self.on_close_clicked)

        self.change_layout_button = builder.get_object("ibus_config_change_layout_button")
        self.change_layout_button.connect("clicked", self.on_change_layout_clicked)
        self.change_layout_button.set_sensitive(self.allows_override)

        self.clear_override_button = builder.get_object("ibus_config_clear_override_button")
        self.clear_override_button.connect("clicked", self.on_clear_override_clicked)

        engine_settings_button = builder.get_object("ibus_config_engine_settings_button")
        if source.preferences:
            engine_settings_button.connect("clicked", self.on_engine_settings_clicked)
        else:
            engine_settings_button.set_visible(False)

        self.update_layout_display()
        self.dialog.show_all()

    def _get_engine_layout(self):
        ibus = IBus.Bus.new()
        if ibus.is_connected():
            engines = ibus.get_engines_by_names([self.source.id])
            if engines:
                return engines[0].get_layout() or "default"
        return "default"

    def _get_engine_layout_display_name(self):
        if not self.engine_layout or self.engine_layout == "default":
            return _("Default")
        got, display_name, short_name, layout, variant = self.xkb_info.get_layout_info(self.engine_layout)
        if got:
            return f"{display_name} ({self.engine_layout})"
        return self.engine_layout

    def run(self):
        return self.dialog.run()

    def destroy(self):
        self.dialog.destroy()

    def on_close_clicked(self, button, data=None):
        self.dialog.response(Gtk.ResponseType.CLOSE)

    def get_current_override(self):
        source_layouts = self.settings.get_value("source-layouts").unpack()
        return source_layouts.get(self.source.id, None)

    def get_layout_display_name(self, layout_id):
        if layout_id is None:
            return None
        got, display_name, short_name, layout, variant = self.xkb_info.get_layout_info(layout_id)
        if got:
            return f"{display_name} ({layout_id})"
        return layout_id

    def update_layout_display(self):
        if not self.allows_override:
            # Engine has a fixed layout requirement
            self.layout_label.set_text(self._get_engine_layout_display_name())
            self.clear_override_button.set_sensitive(False)
            return

        override = self.get_current_override()
        if override:
            display_name = self.get_layout_display_name(override)
            self.layout_label.set_text(display_name)
            self.clear_override_button.set_sensitive(True)
        else:
            self.layout_label.set_text(_("Default"))
            self.clear_override_button.set_sensitive(False)

    def on_change_layout_clicked(self, button, data=None):
        # Show the layout picker in XKB-only mode
        add_dialog = AddKeyboardLayout.AddKeyboardLayoutDialog([], xkb_only=True)
        add_dialog.dialog.set_transient_for(self.dialog)
        add_dialog.dialog.show_all()
        ret = add_dialog.dialog.run()
        if ret == Gtk.ResponseType.OK:
            layout_type, layout_id = add_dialog.response
            self.set_layout_override(layout_id)
        add_dialog.dialog.destroy()

    def on_clear_override_clicked(self, button, data=None):
        self.clear_layout_override()

    def on_engine_settings_clicked(self, button, data=None):
        subprocess.Popen([self.source.preferences], shell=True)

    def set_layout_override(self, layout_id):
        source_layouts = self.settings.get_value("source-layouts").unpack()
        source_layouts[self.source.id] = layout_id
        self.settings.set_value("source-layouts", GLib.Variant("a{ss}", source_layouts))
        self.update_layout_display()

    def clear_layout_override(self):
        source_layouts = self.settings.get_value("source-layouts").unpack()
        if self.source.id in source_layouts:
            del source_layouts[self.source.id]
            self.settings.set_value("source-layouts", GLib.Variant("a{ss}", source_layouts))
        self.update_layout_display()

class LayoutIcon(Gtk.Overlay):
    def __init__(self, file, dupe_id):
        Gtk.Overlay.__init__(self)
        self.file = file
        self.dupe_id = dupe_id

        fi = Gio.FileIcon(file=file)
        flag = Gtk.Image.new_from_gicon(fi, Gtk.IconSize.DND)
        self.add(flag)
        self.drawing = Gtk.DrawingArea(halign=Gtk.Align.FILL, valign=Gtk.Align.FILL)
        self.drawing.connect('draw', self.draw_subscript)
        self.add_overlay(self.drawing)

    def draw_subscript(self, area, cr, data=None):
        if self.dupe_id < 1:
            return

        alloc = area.get_allocation()
        ax, ay, awidth, aheight = (alloc.x, alloc.y, alloc.width, alloc.height)

        x = width = awidth / 2
        y = height = aheight / 2

        cr.set_source_rgba(0, 0, 0, 0.5)
        cr.rectangle(x, y, width, height)
        cr.fill()

        cr.set_source_rgba(1.0, 1.0, 1.0, 0.8)
        cr.rectangle(x + 1, y + 1, width - 2, height - 2)
        cr.fill()

        cr.set_source_rgba(0.0, 0.0, 0.0, 1.0)
        cr.select_font_face("sans", cairo.FontSlant.NORMAL, cairo.FontWeight.BOLD)
        cr.set_font_size(height - 2.0)

        dupe_str = str(self.dupe_id)

        ext = cr.text_extents(dupe_str)
        cr.move_to((x + (width / 2.0) - (ext.width / 2.0)),
                   (y + (height / 2.0) + (ext.height / 2.0)))
        cr.show_text(dupe_str)

class CurrentInputSource(GObject.GObject):
    __gtype_name__ = "CurrentInputSource"
    def __init__(self, item):
        super().__init__()
        self.type, self.id, self.index,         \
            self.display_name, self.short_name, \
            self.flag_name, self.xkbid,         \
            self.xkb_layout, self.xkb_variant,  \
            self.preferences,                   \
            self.dupe_id, self.active           \
                 = item
        if self.preferences is None:
            self.preferences = ''

class CurrentInputSourcesModel(GObject.Object, Gio.ListModel):
    __gtype_name__ = 'CurrentInputSourcesModel'

    def __init__(self):
        super().__init__()
        self._proxy = None
        self._sources = []
        self.interface_settings = Gio.Settings(schema_id="org.cinnamon.desktop.interface")
        self.interface_settings.connect("changed", self.on_interface_settings_changed)
        self.input_source_settings = Gio.Settings(schema_id="org.cinnamon.desktop.input-sources")
        self.input_source_settings.connect("changed::source-layouts", self.on_source_layouts_changed)

        try:
            Gio.DBusProxy.new_for_bus(Gio.BusType.SESSION, Gio.DBusProxyFlags.NONE, None,
                                      "org.Cinnamon", "/org/Cinnamon", "org.Cinnamon", None, self._on_proxy_ready, None)
        except GLib.Error as e:
            print(e.message)
            self._proxy = None

        self.xkb_info = CinnamonDesktop.XkbInfo.new_with_extras()

        self._ibus = IBus.Bus.new()
        if not self._ibus.is_connected():
            print("Connecting to IBus")
            self._ibus.connect("connected", self._on_ibus_connected)
        else:
            print("IBus already connected")
            self._on_ibus_connected(self._ibus)

    @property
    def live(self):
        if self._proxy is None:
            return False
        if self._proxy.get_name_owner() is None:
            return False
        return True

    def _get_layout_override(self, engine_id):
        source_layouts = self.input_source_settings.get_value("source-layouts").unpack()
        return source_layouts.get(engine_id, None)

    def _get_layout_display_name(self, layout_id):
        got, display_name, short_name, layout, variant = self.xkb_info.get_layout_info(layout_id)
        if got:
            return display_name
        return layout_id

    def _on_ibus_connected(self, ibus, data=None):
        if self._proxy is None:
            self.refresh_input_source_list()

    def _on_proxy_ready(self, obj, result, data=None):
        try:
            self._proxy = Gio.DBusProxy.new_for_bus_finish(result)
            self._proxy.connect("g-signal", self._on_proxy_signal)
            self._proxy.connect("notify::g-name-owner", self._on_cinnamon_state_changed)
            self.refresh_input_source_list()
        except GLib.Error as e:
            print(f"Keyboard module could not establish proxy for org.Cinnamon: {e}")

    def _on_proxy_signal(self, proxy, sender_name, signal_name, parameters):
        if signal_name == "InputSourcesChanged":
            self.refresh_input_source_list()

    def _on_cinnamon_state_changed(self, proxy, pspec, data=None):
        # If Cinnamon crashes, this will happen, reload our sources using xkb and ibus.
        # This isn't reliable to detect Cinnamon restarting, as it will regain a name-owner
        # before org.Cinnamon is exported, causing GetInputSources to fail.
        #
        # Fortunately, InputSourcesChanged will fire at startup, which is perfect to re-sync
        # with Cinnamon.
        if proxy.get_name_owner() is None:
            self.refresh_input_source_list()

    def on_interface_settings_changed(self, settings, key, data=None):
        if key.startswith("keyboard-layout-"):
            self.refresh_input_source_list()

    def on_source_layouts_changed(self, settings, key, data=None):
        self.refresh_input_source_list()

    def refresh_input_source_list(self):
        if self.live:
            layouts = self._proxy.GetInputSources()
        else:
            sources = self.input_source_settings.get_value("sources")
            layouts = []
            index = 0

            for type_, id_ in sources:
                if type_ == "xkb":
                    got, display_name, short_name, layout, variant = self.xkb_info.get_layout_info(id_)
                    if got:
                        layouts.append(
                            (type_, id_, index, display_name,
                             None, None, id_,
                             layout, variant, None,
                             0, False)
                        )
                else:
                    engines = self._ibus.get_engines_by_names([id_])
                    if len(engines) > 0:
                        engine = engines[0]
                        display_name = AddKeyboardLayout.make_ibus_display_name(engine)
                        layouts.append(
                            (type_, id_, index, display_name,
                             None, None, id_,
                             engine.get_layout(), engine.get_layout_variant(), engine.get_setup(),
                             0, False)
                        )

                index += 1

        old_layouts = self._sources
        new_layouts = []

        for layout in layouts:
            new_layouts.append(CurrentInputSource(layout))

        self._sources = new_layouts
        self.column_size_group = Gtk.SizeGroup(Gtk.SizeGroupMode.BOTH)
        self.items_changed(0, len(old_layouts), len(new_layouts))

    def show_add_layout_dialog(self):
        used_ids = [source.id for source in self._sources]
        add_dialog = AddKeyboardLayout.AddKeyboardLayoutDialog(used_ids)
        add_dialog.dialog.show_all()
        ret = add_dialog.dialog.run()
        if ret == Gtk.ResponseType.OK:
            self.add_layout(*add_dialog.response)
        add_dialog.dialog.destroy()

    def create_row(self, source, data=None):
        row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        # Build display name, including layout override for IBus sources
        display_name = GLib.markup_escape_text(source.display_name)
        if source.type == "ibus":
            layout_override = self._get_layout_override(source.id)
            if layout_override:
                layout_display = GLib.markup_escape_text(self._get_layout_display_name(layout_override))
                markup = f"<b>{display_name}</b>  <small>|  {layout_display}</small>"
            else:
                markup = f"<b>{display_name}</b>"
        else:
            markup = f"<b>{display_name}</b>"
        label = Gtk.Label(label=markup, xalign=0.0, use_markup=True, margin_start=4)
        row.pack_start(label, True, True, 0)

        if self.live:
            indicator_done = False

            if self.interface_settings.get_boolean("keyboard-layout-show-flags"):
                flag_file = f"/usr/share/iso-flag-png/{source.flag_name}.png"
                if os.path.exists(flag_file):
                    file = Gio.File.new_for_path(flag_file)
                    flag = LayoutIcon(file, source.dupe_id)

                    self.column_size_group.add_widget(flag)
                    row.pack_start(flag, False, False, 0)
                    indicator_done = True

            if not indicator_done:
                markup = f"<span size='x-large' weight='bold'>{source.short_name}</span>"
                label = Gtk.Label(xalign=0.0, label=markup, use_markup=True)
                self.column_size_group.add_widget(label)
                row.pack_end(label, False, False, 0)
        else:
            dummy_markup = "<span size='x-large' weight='bold'> </span>"
            label = Gtk.Label(xalign=0.0, label=dummy_markup, use_markup=True)
            self.column_size_group.add_widget(label)
            row.pack_end(label, False, False, 0)

        row.show_all()
        row.input_source = source
        return row

    def do_get_item(self, position):
        return self._sources[position]

    def do_get_item_type(self):
        return CurrentInputSource

    def do_get_n_items(self):
        return len(self._sources)

    def get_item_index(self, item):
        if item not in self._sources:
            return -1
        return self._sources.index(item)

    def add_layout(self, type_, layout_id):
        raw_sources = self.input_source_settings.get_value("sources")
        new_sources = []

        # raw_sources is a variant, not a list.
        for source_info in raw_sources:
            new_sources.append(source_info)

        new_sources.append((type_, layout_id))
        self.input_source_settings.set_value("sources", GLib.Variant("a(ss)", new_sources))

    def remove_layout(self, source):
        raw_sources = self.input_source_settings.get_value("sources")
        new_sources = []

        # shouldn't happen but here we're not relying on cinnamon, but dconf.
        if len(raw_sources) == 1:
            print("Cannot remove the only remaining layout!!")
            return

        for source_info in raw_sources:
            type_, id_ = source_info
            if source.id == id_ and source.type == type_:
                continue
            new_sources.append(source_info)

        self.input_source_settings.set_value("sources", GLib.Variant("a(ss)", new_sources))

    def move_layout_up(self, source):
        raw_sources = self.input_source_settings.get_value("sources")
        new_sources = []

        # raw_sources is a variant, not a list.
        for source_info in raw_sources:
            new_sources.append(source_info)

        t = (source.type, source.id)
        try:
            idx = new_sources.index(t)
            if idx == 0:
                return
            element = new_sources.pop(idx)
            new_sources.insert(idx - 1, element)
            self.input_source_settings.set_value("sources", GLib.Variant("a(ss)", new_sources))
        except Exception as e:
            print("Could not move layout", e)

    def move_layout_down(self, source):
        raw_sources = self.input_source_settings.get_value("sources")
        new_sources = []

        # raw_sources is a variant, not a list.
        for source_info in raw_sources:
            new_sources.append(source_info)

        t = (source.type, source.id)
        try:
            idx = new_sources.index(t)
            if idx == len(new_sources) - 1:
                return
            element = new_sources.pop(idx)
            new_sources.insert(idx + 1, element)
            self.input_source_settings.set_value("sources", GLib.Variant("a(ss)", new_sources))
        except Exception as e:
            print("Could not move layout", e)
